#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csscs script.
#
# Copyright (C) 2020 Luigi Tibaldo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import cscripts
from testing import test


# =========================== #
# Test class for csscs script #
# =========================== #
class Test(test):
    """
    Test class for csscs script

    This test class makes unit tests for the csscs script by using it from
    the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test datasets and parameters
        self._evs_offaxis = self._datadir + '/crab_offaxis1.fits'
        self._exclusion   = self._datadir + '/crab_exclusion.fits'
        self._expcube     = self._datadir + '/crab_expcube.fits'
        self._psfcube     = self._datadir + '/crab_psfcube.fits'
        self._bkgcube     = self._datadir + '/crab_bkgcube.fits'
        self._bkgmodel    = self._datadir + '/crab_bkgcube.xml'
        self._edispcube   = self._datadir + '/crab_edispcube.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csscs')

        # Append tests
        self.append(self._test_cmd, 'Test csscs on command line')
        self.append(self._test_python, 'Test csscs from Python')
        self.append(self._test_pickeling, 'Test csscs pickeling')

        # Return
        return

    # Test csscs on command line
    def _test_cmd(self):
        """
        Test csscs on the command line
        """
        # Set script name
        csscs = self._script('csscs')

        # Setup csscs command
        cmd = csscs + ' inobs="' + self._cntcube + '"' + \
              ' expcube="' + self._expcube + '"' + \
              ' psfcube="' + self._psfcube + '"' + \
              ' bkgcube="' + self._bkgcube + '"' + \
              ' inmodel="' + self._bkgmodel + '"' + \
              ' srcnames="Crab"' + \
              ' emin=1. emax=100.0 nxpix=2 nypix=2' + \
              ' binsz=0.1 coordsys="CEL" proj="CAR"' + \
              ' xref=83.63 yref=22.01 rad=0.2' + \
              ' outfile="csscs_cmd1.fits"' + \
              ' logfile="csscs_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_value(self._execute(cmd), 0,
                        'Check successful execution from command line')

        # Setup csscs command
        cmd = csscs + ' inobs="' + self._cntcube + '"' + \
              ' expcube="' + self._expcube + '"' + \
              ' psfcube="' + self._psfcube + '"' + \
              ' bkgcube="' + self._bkgcube + '"' + \
              ' inmodel="' + self._bkgmodel + '"' + \
              ' srcnames="Crab;source_that_is_not_in_the_model"' + \
              ' emin=1. emax=100.0 nxpix=2 nypix=2' + \
              ' binsz=0.1 coordsys="CEL" proj="CAR"' + \
              ' xref=83.63 yref=22.01 rad=0.2' + \
              ' outfile="csscs_cmd1.fits"' + \
              ' logfile="csscs_cmd1.log" chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
                         'Check invalid source name when executed from command line')

        # Check csscs --help
        self._check_help(csscs)

        # Return
        return

    # Test csscs from Python
    def _test_python(self):
        """
        Test csscs from Python
        """
        # Allocate csscs
        script = cscripts.csscs()

        # Check that saving does not nothing
        script['logfile'] = 'csscs_py0.log'
        script['outfile'] = 'csscs_py0.fits'
        script.logFileOpen()
        script.save()
        self.test_assert(not os.path.isfile('csscs_py0.fits'),
                         'Check that no FITS file has been created')

        # Check that clearing does not lead to an exception or segfault
        script.clear()

        # Set up csscs for simple binned analysis
        script['logfile'] = 'csscs_py1.log'
        script['inobs'] = self._cntcube
        script['expcube'] = self._expcube
        script['psfcube'] = self._psfcube
        script['bkgcube'] = self._bkgcube
        script['inmodel'] = self._bkgmodel
        script['srcnames'] = 'Crab'
        script['emin'] = 1.
        script['emax'] = 100.
        script['nxpix'] = 2
        script['nypix'] = 2
        script['binsz'] = 0.1
        script['coordsys'] = 'CEL'
        script['proj'] = 'CAR'
        script['xref'] = 83.63
        script['yref'] = 22.01
        script['rad'] = 0.2
        script['outfile'] = 'csscs_py1.fits'
        script['chatter'] = 4

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py1.fits')

        # Set up for same test without TS computation
        script = cscripts.csscs()
        script['logfile']  = 'csscs_py2.log'
        script['inobs']    = self._cntcube
        script['expcube']  = self._expcube
        script['psfcube']  = self._psfcube
        script['bkgcube']  = self._bkgcube
        script['inmodel']  = self._bkgmodel
        script['srcnames'] = 'Crab'
        script['emin']     = 1.
        script['emax']     = 100.
        script['nxpix']    = 2
        script['nypix']    = 2
        script['binsz']    = 0.1
        script['coordsys'] = 'CEL'
        script['proj']     = 'CAR'
        script['xref']     = 83.63
        script['yref']     = 22.01
        script['rad']      = 0.2
        script['calc_ts']  = False
        script['outfile']  = 'csscs_py2.fits'
        script['chatter']  = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py2.fits', calc_ts=False)

        # # Set up for same test with upper limit computation
        script = cscripts.csscs()
        script['logfile']   = 'csscs_py3.log'
        script['inobs']     = self._cntcube
        script['expcube']   = self._expcube
        script['psfcube']   = self._psfcube
        script['bkgcube']   = self._bkgcube
        script['inmodel']   = self._bkgmodel
        script['srcnames']  = 'Crab'
        script['emin']      = 1.
        script['emax']      = 100.
        script['nxpix']     = 2
        script['nypix']     = 2
        script['binsz']     = 0.1
        script['coordsys']  = 'CEL'
        script['proj']      = 'CAR'
        script['xref']      = 83.63
        script['yref']      = 22.01
        script['rad']       = 0.2
        script['calc_ts']   = True
        script['calc_ulim'] = True
        script['outfile']   = 'csscs_py3.fits'
        script['chatter']   = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py3.fits', calc_ulim=True)

        # # Check that results can be piped in memory
        self._check_results_mem(script, calc_ulim=True)

        # Set up for test without multiprocessing
        script = cscripts.csscs()
        script['logfile']  = 'csscs_py4.log'
        script['inobs']    = self._cntcube
        script['expcube']  = self._expcube
        script['psfcube']  = self._psfcube
        script['bkgcube']  = self._bkgcube
        script['inmodel']  = self._bkgmodel
        script['srcnames'] = 'Crab'
        script['emin']     = 1.
        script['emax']     = 100.
        script['nxpix']    = 2
        script['nypix']    = 2
        script['binsz']    = 0.1
        script['coordsys'] = 'CEL'
        script['proj']     = 'CAR'
        script['xref']     = 83.63
        script['yref']     = 22.01
        script['rad']      = 0.2
        script['outfile']  = 'csscs_py4.fits'
        script['nthreads'] = 1
        script['chatter']  = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py4.fits')

        # Set up for test with energy dispersion
        script = cscripts.csscs()
        script['logfile']   = 'csscs_py5.log'
        script['inobs']     = self._cntcube
        script['expcube']   = self._expcube
        script['psfcube']   = self._psfcube
        script['bkgcube']   = self._bkgcube
        script['edisp']     = True
        script['edispcube'] = self._edispcube
        script['inmodel']   = self._bkgmodel
        script['srcnames']  = 'Crab'
        script['emin']      = 1.
        script['emax']      = 100.
        script['nxpix']     = 2
        script['nypix']     = 2
        script['binsz']     = 0.1
        script['coordsys']  = 'CEL'
        script['proj']      = 'CAR'
        script['xref']      = 83.63
        script['yref']      = 22.01
        script['rad']       = 0.2
        script['outfile']   = 'csscs_py5.fits'
        script['chatter']   = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py5.fits')

        # Set up for test in On/Off mode with background model
        script = cscripts.csscs()
        script['logfile']     = 'csscs_py6.log'
        script['inobs']       = self._evs_offaxis
        script['caldb']       = self._caldb
        script['irf']         = self._irf
        script['inmodel']     = self._model
        script['srcnames']    = 'Crab'
        script['emin']        = 1.
        script['emax']        = 100.
        script['nxpix']       = 2
        script['nypix']       = 2
        script['binsz']       = 0.1
        script['coordsys']    = 'CEL'
        script['proj']        = 'CAR'
        script['xref']        = 83.63
        script['yref']        = 22.01
        script['rad']         = 0.2
        script['method']      = 'ONOFF'
        script['inexclusion'] = self._exclusion
        script['enumbins']    = 2
        script['outfile']     = 'csscs_py6.fits'
        script['chatter']     = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py6.fits')

        # Same test with exclusion map piped in memory
        script = cscripts.csscs()
        script['logfile']  = 'csscs_py7.log'
        script['inobs']    = self._evs_offaxis
        script['caldb']    = self._caldb
        script['irf']      = self._irf
        script['inmodel']  = self._model
        script['srcnames'] = 'Crab'
        script['emin']     = 1.
        script['emax']     = 100.
        script['nxpix']    = 2
        script['nypix']    = 2
        script['binsz']    = 0.1
        script['coordsys'] = 'CEL'
        script['proj']     = 'CAR'
        script['xref']     = 83.63
        script['yref']     = 22.01
        script['rad']      = 0.2
        script['method']   = 'ONOFF'
        script['enumbins'] = 2
        script['outfile']  = 'csscs_py7.fits'
        script['chatter']  = 2

        # Set exclusion map
        exclmap = gammalib.GSkyRegionMap(self._exclusion)
        script.exclusion_map(exclmap)

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py7.fits')

        # Set up for test in On/Off mode without background model
        script = cscripts.csscs()
        script['logfile']       = 'csscs_py8.log'
        script['inobs']         = self._evs_offaxis
        script['caldb']         = self._caldb
        script['irf']           = self._irf
        script['inmodel']       = self._model
        script['srcnames']      = 'Crab'
        script['emin']          = 1.
        script['emax']          = 100.
        script['nxpix']         = 2
        script['nypix']         = 2
        script['binsz']         = 0.1
        script['coordsys']      = 'CEL'
        script['proj']          = 'CAR'
        script['xref']          = 83.63
        script['yref']          = 22.01
        script['rad']           = 0.2
        script['method']        = 'ONOFF'
        script['use_model_bkg'] = False
        script['inexclusion']   = self._exclusion
        script['enumbins']      = 2
        script['outfile']       = 'csscs_py8.fits'
        script['chatter']       = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py8.fits')

        # Set up for test in unbinned mode
        script = cscripts.csscs()
        script['logfile']  = 'csscs_py9.log'
        script['inobs']    = self._evs_offaxis
        script['caldb']    = self._caldb
        script['irf']      = self._irf
        script['inmodel']  = self._model
        script['srcnames'] = 'Crab'
        script['emin']     = 1.
        script['emax']     = 100.
        script['nxpix']    = 2
        script['nypix']    = 2
        script['binsz']    = 0.1
        script['coordsys'] = 'CEL'
        script['proj']     = 'CAR'
        script['xref']     = 83.63
        script['yref']     = 22.01
        script['rad']      = 0.2
        script['method']   = 'UNBINNED'
        script['enumbins'] = 2
        script['outfile']  = 'csscs_py9.fits'
        script['chatter']  = 2

        # Run csscs script and save results
        script.logFileOpen()
        script.execute()

        # Check result
        self._check_result_file('csscs_py9.fits')

        # Return
        return

    # Test csscs pickeling
    def _test_pickeling(self):
        """
        Test csscs pickeling
        """
        # Perform pickeling test of empty class
        self._pickeling(cscripts.csscs())

        # Setup script for pickling text
        script = cscripts.csscs()
        script['logfile']  = 'csscs_py1_pickle.log'
        script['inobs']    = self._cntcube
        script['expcube']  = self._expcube
        script['psfcube']  = self._psfcube
        script['bkgcube']  = self._bkgcube
        script['inmodel']  = self._bkgmodel
        script['srcnames'] = 'Crab'
        script['emin']     = 1.
        script['emax']     = 100.
        script['nxpix']    = 2
        script['nypix']    = 2
        script['binsz']    = 0.1
        script['coordsys'] = 'CEL'
        script['proj']     = 'CAR'
        script['xref']     = 83.63
        script['yref']     = 22.01
        script['rad']      = 0.2
        script['outfile']  = 'csscs_py1_pickle.fits'
        script['chatter']  = 2

        # Perform pickeling tests of filled class
        obj = self._pickeling(script)

        # Run csscs script and save result
        obj.logFileOpen()  # Make sure we get a log file
        obj.execute()

        # Check result file
        self._check_result_file('csscs_py1_pickle.fits')

        # Return
        return

    # Check access to results in memory'
    def _check_results_mem(self, script, srcnames=['Crab'], nx=2, ny=2, calc_ts=True,
                           calc_ulim=False):
        """
        Check results in memory

        Parameters
        ----------
        script : `~cscripts.csscs`
            csscs object
        srcnames : list, optional
            Source names
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        calc_ts : bool, optional
            Check TS HDU
        calc_ulim : bool, optional
            Check upper limit HDU
        """
        # Check that the entire fits container can be extracted
        fits = script.fits()

        # Dump to disk and verify
        fits.saveto('csscs_py3_mem.fits', True)
        self._check_result_file('csscs_py3_mem.fits', srcnames=srcnames, nx=nx, ny=ny,
                                calc_ts=calc_ts, calc_ulim=calc_ulim)

        # Now check direct access in memory for skymaps
        # Loop over sources
        for name in srcnames:
            # Check flux map
            skymap = script.flux(name)
            self._check_map(skymap, nx=nx, ny=ny)
            # Check flux error map
            skymap = script.flux_error(name)
            self._check_map(skymap, nx=nx, ny=ny)
            # If requested check TS map
            if calc_ts:
                skymap = script.ts(name)
                self._check_map(skymap, nx=nx, ny=ny)
            # If requested check upper limit map
            if calc_ulim:
                skymap = script.ulimit(name)
                self._check_map(skymap, nx=nx, ny=ny)

        return

    # Check result file
    def _check_result_file(self, filename, srcnames=['Crab'], nx=2, ny=2,
                           calc_ts=True, calc_ulim=False):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Sky map file name
        srcnames : list, optional
            Source names
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        calc_ts : bool, optional
            Check TS HDU
        calc_ulim : bool, optional
            Check upper limit HDU
        """
        # Loop over source names
        for name in srcnames:
            # Check flux map
            skymap = gammalib.GSkyMap(filename + '[' + name + ' FLUX]')
            self._check_map(skymap, nx=nx, ny=ny)
            # Check flux error map
            skymap = gammalib.GSkyMap(filename + '[' + name + ' FLUX ERROR]')
            self._check_map(skymap, nx=nx, ny=ny)
            # If requested check TS map
            if calc_ts:
                skymap = gammalib.GSkyMap(filename + '[' + name + ' TS]')
                self._check_map(skymap, nx=nx, ny=ny)
            # If requested check upper limit map
            if calc_ulim:
                skymap = gammalib.GSkyMap(filename + '[' + name + ' FLUX UPPER LIMIT]')
                self._check_map(skymap, nx=nx, ny=ny)

        # Return
        return

    # Check sky map
    def _check_map(self, skymap, nx=2, ny=2):
        """
        Check sky map

        Parameters
        ----------
        skymap : `~gammalib.GSkyMap`
            Sky map
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Check dimensions
        self.test_value(skymap.nmaps(), 1, 'Check number of maps')
        self.test_value(skymap.nx(), nx, 'Check for number of X pixels')
        self.test_value(skymap.ny(), ny, 'Check for number of Y pixels')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Generate IRFs in CALDB format from a ROOT offaxis performance file
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import sys
import math
import copy
from datetime import datetime
import gammalib
import ctools
from cscripts import calutils


# ================== #
# csroot2caldb class #
# ================== #
class csroot2caldb(ctools.cscript):
    """
    Generates IRFs in CALDB from a ROOT offaxis performance file
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Query parameters
        self['infile'].filename()
        self['outdir'].string()
        self['inst'].string()
        self['id'].string()
        self['analysis'].string()
        self['zenith'].real()
        self['azimuth'].real()
        self['version'].string()
        self['psftype'].string()
        self['split'].boolean()
        self['norm1d'].boolean()
        self['rebin'].boolean()
        self['eascale'].real()
        self['bgdscale'].real()
        self['bgdoversample'].integer()
        self['bgdinfill'].boolean()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    def _init_metadata(self):
        """
        Initialise dictionary containing the IRF metadata

        Returns
        -------
        irf : dict
            IRF metadata dictionary
        """
        # Initialise IRF metadata dictionary
        irf = {}

        # Set calibration information
        cal_name     = 'NAME('+self['id'].string()+')'
        cal_version  = 'VERSION('+self['version'].string()+')'
        cal_cut      = 'CLASS(BEST)'
        cal_analysis = 'ANALYSIS('+self['analysis'].string()+')'
        cal_zenith   = 'ZENITH(%.3f)deg' % self['zenith'].real()
        cal_azimuth  = 'AZIMUTH(%.3f)deg' % self['azimuth'].real()
        cal_bounds   = [cal_name, cal_version, cal_cut, cal_analysis, \
                        cal_zenith, cal_azimuth]

        # Set IRF information
        irf['CAL_TEL']        = 'CTA'
        irf['CAL_INST']       = self['inst'].string().upper()
        irf['CAL_OBSID']      = self['id'].string()
        irf['CAL_DET']        = 'NONE'
        irf['CAL_FLT']        = 'NONE'
        irf['CAL_CLASS']      = 'BCF'
        irf['CAL_TYPE']       = 'DATA'
        irf['CAL_QUAL']       = 0            # 0=good, 1=bad, 2=dubious, ...
        irf['CAL_DATE']       = '14/01/30'
        irf['VAL_DATE']       = '2014-01-30'
        irf['VAL_TIME']       = '00:00:00'
        irf['REF_TIME']       = 51544.0
        irf['EA_NAME']        = 'EFF_AREA'
        irf['EA_HDUCLAS3']    = 'FULL-ENCLOSURE'
        irf['EA_HDUCLAS4']    = '2D'
        irf['EA_DOC']         = 'CAL/GEN/92-019'
        irf['EA_BOUNDS']      = copy.deepcopy(cal_bounds)
        irf['EA_DESC']        = 'CTA effective area'
        irf['PSF_NAME']       = 'RPSF'
        irf['PSF_HDUCLAS3']   = 'FULL-ENCLOSURE'
        irf['PSF_HDUCLAS4']   = '3GAUSS'
        irf['PSF_DOC']        = 'CAL/GEN/92-020'
        irf['PSF_BOUNDS']     = copy.deepcopy(cal_bounds)
        irf['PSF_DESC']       = 'CTA point spread function'
        irf['EDISP_NAME']     = 'EDISP'
        irf['EDISP_HDUCLAS3'] = 'FULL-ENCLOSURE'
        irf['EDISP_HDUCLAS4'] = '2D'
        irf['EDISP_DOC']      = '???'
        irf['EDISP_BOUNDS']   = copy.deepcopy(cal_bounds)
        irf['EDISP_DESC']     = 'CTA energy dispersion'
        irf['BKG_NAME']       = 'BKG'
        irf['BKG_HDUCLAS3']   = 'FULL-ENCLOSURE'
        irf['BKG_HDUCLAS4']   = '3D'
        irf['BKG_DOC']        = '???'
        irf['BKG_BOUNDS']     = copy.deepcopy(cal_bounds)
        irf['BKG_DESC']       = 'CTA background'

        # Set parameter dependent IRF information
        if self['psftype'].string() == 'King':
            irf['PSF_HDUCLAS4'] = 'KING'
        else:
            irf['PSF_HDUCLAS4'] = '3GAUSS'

        # Return metadata
        return irf

    def _make_dirs(self, version, irf):
        """
        Generate CALDB directory structure for one observation identifier

        The structure is given by

            data/<tel>/<inst>/bcf/<obsid>

        where <tel> is "cta" and <inst> is the instrument specified in the
        CALDB constructor (the instrument may be used for different array
        configurations).

        Parameters
        ----------
        version : str
            Version string
        irf : dict
            IRF metadata dictionary

        Returns
        -------
        ds : dict
            Directory structure information
        """
        # Initialise directory structure dictionary
        ds = {}

        # Set base directory
        base_dir  = 'data'
        base_dir += '/'+irf['CAL_TEL'].lower()
        base_dir += '/'+irf['CAL_INST'].lower()

        # Set directory names
        obs_dir          = base_dir+'/bcf/'+irf['CAL_OBSID']
        ds['EA_DIR']    = obs_dir
        ds['PSF_DIR']   = obs_dir
        ds['EDISP_DIR'] = obs_dir
        ds['BKG_DIR']   = obs_dir

        # Set path to response components. If the "outdir" parameter is set
        # then prefix the value to the path of the response components
        outdir = self['outdir'].string()
        if len(outdir) > 0:
            ds['BASE_PATH']  = outdir+'/'+base_dir
            ds['EA_PATH']    = outdir+'/'+ds['EA_DIR']
            ds['PSF_PATH']   = outdir+'/'+ds['PSF_DIR']
            ds['EDISP_PATH'] = outdir+'/'+ds['EDISP_DIR']
            ds['BKG_PATH']   = outdir+'/'+ds['BKG_DIR']
        else:
            ds['BASE_PATH']  = base_dir
            ds['EA_PATH']    = ds['EA_DIR']
            ds['PSF_PATH']   = ds['PSF_DIR']
            ds['EDISP_PATH'] = ds['EDISP_DIR']
            ds['BKG_PATH']   = ds['BKG_DIR']

        # Set IRF component file names. If the "split" parameter is "yes" then
        # split the IRF components over several files.
        if self['split'].boolean():
            ds['EA_FILE']    = 'ea_'+version+'.fits'
            ds['PSF_FILE']   = 'psf_'+version+'.fits'
            ds['EDISP_FILE'] = 'edisp_'+version+'.fits'
            ds['BKG_FILE']   = 'bgd_'+version+'.fits'
        else:
            ds['EA_FILE']    = 'irf_'+version+'.fits'
            ds['PSF_FILE']   = 'irf_'+version+'.fits'
            ds['EDISP_FILE'] = 'irf_'+version+'.fits'
            ds['BKG_FILE']   = 'irf_'+version+'.fits'

        # Create directory structure
        if not os.path.isdir(ds['EA_PATH']):
            os.makedirs(ds['EA_PATH'])
        if not os.path.isdir(ds['PSF_PATH']):
            os.makedirs(ds['PSF_PATH'])
        if not os.path.isdir(ds['EDISP_PATH']):
            os.makedirs(ds['EDISP_PATH'])
        if not os.path.isdir(ds['BKG_PATH']):
            os.makedirs(ds['BKG_PATH'])

        # Return ds
        return ds

    def _open(self, irf, ds):
        """
        Open existing or create new calibration

        The actual version will put all calibrations in the same file, although
        each part of the response function will have its own logical name. We
        can thus easily modify the script to put each calibration information
        in a separate file

        Parameters
        ----------
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Open calibration database index
        ds['CIF'] = gammalib.GFits(ds['BASE_PATH']+'/caldb.indx', True)

        # If file has no CIF extension than create it now
        if not ds['CIF'].contains('CIF'):
            ds['CIF'].append(calutils.create_cif_table())
        ds['HDU_CIF'] = ds['CIF'].table('CIF')

        # Set IRF component filenames
        ea_filename    = ds['EA_PATH']+'/'+ds['EA_FILE']
        psf_filename   = ds['PSF_PATH']+'/'+ds['PSF_FILE']
        edisp_filename = ds['EDISP_PATH']+'/'+ds['EDISP_FILE']
        bgd_filename   = ds['BKG_PATH']+'/'+ds['BKG_FILE']

        # Open IRF component files
        if self['split'].boolean():
            ds['EA_FITS']    = gammalib.GFits(ea_filename, True)
            ds['PSF_FITS']   = gammalib.GFits(psf_filename, True)
            ds['EDISP_FITS'] = gammalib.GFits(edisp_filename, True)
            ds['BKG_FITS']   = gammalib.GFits(bgd_filename, True)
        else:
            ds['EA_FITS']    = gammalib.GFits(ea_filename, True)
            ds['PSF_FITS']   = ds['EA_FITS']
            ds['EDISP_FITS'] = ds['EA_FITS']
            ds['BKG_FITS']   = ds['EA_FITS']

        # Open HDUs
        ds['HDU_EA']    = self._open_hdu(ds['EA_FITS'], "EFFECTIVE AREA",
                                         irf['EA_NAME'], irf['EA_HDUCLAS3'],
                                         irf['EA_HDUCLAS4'], irf['EA_DOC'],
                                         irf)
        ds['HDU_PSF']   = self._open_hdu(ds['PSF_FITS'], "POINT SPREAD FUNCTION",
                                         irf['PSF_NAME'], irf['PSF_HDUCLAS3'],
                                         irf['PSF_HDUCLAS4'], irf['PSF_DOC'],
                                         irf)
        ds['HDU_EDISP'] = self._open_hdu(ds['EDISP_FITS'], "ENERGY DISPERSION",
                                         irf['EDISP_NAME'], irf['EDISP_HDUCLAS3'],
                                         irf['EDISP_HDUCLAS4'], irf['EDISP_DOC'],
                                         irf)
        ds['HDU_BKG']   = self._open_hdu(ds['BKG_FITS'], "BACKGROUND",
                                         irf['BKG_NAME'], irf['BKG_HDUCLAS3'],
                                         irf['BKG_HDUCLAS4'], irf['BKG_DOC'],
                                         irf)

        # Return
        return

    def _close(self, irf, ds):
        """
        Close calibration FITS files

        Parameters
        ----------
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Add information to CIF. We do this now as before we did not
        # necessarily have all the information at hand (in particular about
        # the boundaries)
        self._add_cif_info(irf, ds)

        # Save and close CIF
        ds['CIF'].save(True)
        ds['CIF'].close()

        # Close all IRF components. If the files have been split all components
        # need to be closed, otherwise only the effective area component needs
        # to be closed as representative for all components.
        ds['EA_FITS'].save(True)
        ds['EA_FITS'].close()
        if self['split'].boolean():
            ds['PSF_FITS'].save(True)
            ds['EDISP_FITS'].save(True)
            ds['BKG_FITS'].save(True)
            ds['PSF_FITS'].close()
            ds['EDISP_FITS'].close()
            ds['BKG_FITS'].close()

        # Return
        return

    def _open_hdu(self, fits, extname, name, hduclas3, hduclas4, doc, irf):
        """
        Open HDU

        Opens a FITS binary table with given "extname". If HDU does not exist
        in the FITS file it will be created and appended to the FITS file.

        Parameters
        ----------
        fits : `~gammalib.GFits`
            FITS file
        extname : str
            Extension name
        name : str
            Name string
        hduclas3 : str
            HDU class 3 string
        hduclas4 : str
            HDU class 4 string
        doc : str
            Document string
        irf : dict
            IRF metadata dictionary

        Returns
        -------
        table : `~gammalib.GFitsBinTable`
            FITS binary table
        """
        # Create table if it does not yet exist
        if not fits.contains(extname):

            # Create binary table
            table = gammalib.GFitsBinTable()

            # Set extension name
            table.extname(extname)

            # Set OGIP keywords
            self._set_ogip_keywords(table, doc,
                                    ['RESPONSE', name, hduclas3, hduclas4],
                                    irf)

            # Append table to FITS file
            fits.append(table)

        # Return FITS table
        return fits.table(extname)

    def _set_ogip_keywords(self, hdu, hdudoc, hduclas, irf):
        """
        Set standard OGIP keywords for extension

        Parameters
        ----------
        hdu : `~gammalig.GFitsHDU`
            Header Data Unit
        hdudoc : str
            Documentation reference string
        hduclas : list of str
            List of HDUCLAS fields
        irf : dict
            IRF dictonary
        """
        # Set UTC date of file creation
        utc = datetime.utcnow().isoformat()[:19]

        # Set keywords
        hdu.card('ORIGIN', 'IRAP', 'Name of organization making this file')
        hdu.card('DATE', utc, 'File creation date (YYYY-MM-DDThh:mm:ss UTC)')
        hdu.card('TELESCOP', irf['CAL_TEL'], 'Name of telescope')
        hdu.card('INSTRUME', irf['CAL_INST'], 'Name of instrument')
        hdu.card('DETNAM', irf['CAL_DET'], 'Name of detector')
        hdu.card('HDUCLASS', 'OGIP', 'HDU class')
        hdu.card('HDUDOC', hdudoc, 'HDU documentation')
        for i, item in enumerate(hduclas):
            key = 'HDUCLAS%d' % (i+1)
            hdu.card(key, item, 'HDU class')
        hdu.card('HDUVERS', '0.2', 'HDU version')

        # Return
        return

    def _add_cif_info(self, irf, ds):
        """
        Add information to CIF extension

        Parameters
        ----------
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Set list of IRF component names
        names = ['EA', 'PSF', 'EDISP', 'BKG']

        # Initialise CIF row index
        row = ds['HDU_CIF'].nrows()

        # Append rows for all components to CIF extension
        ds['HDU_CIF'].append_rows(len(names))

        # Add information for all components
        for name in names:

            # Set dictionary keys for this component
            key_dir    = '%s_DIR' % name
            key_file   = '%s_FILE' % name
            key_name   = '%s_NAME' % name
            key_desc   = '%s_DESC' % name
            key_bounds = '%s_BOUNDS' % name

            # Set generic information
            ds['HDU_CIF']['TELESCOP'][row] = irf['CAL_TEL']
            ds['HDU_CIF']['INSTRUME'][row] = irf['CAL_INST']
            ds['HDU_CIF']['DETNAM'][row]   = irf['CAL_DET']
            ds['HDU_CIF']['FILTER'][row]   = irf['CAL_FLT']
            ds['HDU_CIF']['CAL_DEV'][row]  = 'ONLINE'
            ds['HDU_CIF']['CAL_CLAS'][row] = irf['CAL_CLASS']
            ds['HDU_CIF']['CAL_DTYP'][row] = irf['CAL_TYPE']
            ds['HDU_CIF']['CAL_VSD'][row]  = irf['VAL_DATE']
            ds['HDU_CIF']['CAL_VST'][row]  = irf['VAL_TIME']
            ds['HDU_CIF']['REF_TIME'][row] = irf['REF_TIME']
            ds['HDU_CIF']['CAL_QUAL'][row] = irf['CAL_QUAL']
            ds['HDU_CIF']['CAL_DATE'][row] = irf['CAL_DATE']

            # Set component specific information
            ds['HDU_CIF']['CAL_DIR'][row]   = ds[key_dir]
            ds['HDU_CIF']['CAL_FILE'][row]  = ds[key_file]
            ds['HDU_CIF']['CAL_CNAM'][row]  = irf[key_name]
            ds['HDU_CIF']['CAL_DESC'][row]  = irf[key_desc]
            ds['HDU_CIF']['CAL_XNO'][row]   = 1
            for i in range(9):
                if i >= len(irf[key_bounds]):
                    ds['HDU_CIF']['CAL_CBD'][row,i] = 'NONE'
                else:
                    ds['HDU_CIF']['CAL_CBD'][row,i] = irf[key_bounds][i]

            # Increment row index
            row += 1

        # Return
        return

    def _set_cif_keywords(self, hdu, name, bounds, desc, irf):
        """
        Set standard CIF keywords for extension

        Parameters
        ----------
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Calibration name
        bounds : list of str
            Calibration boundaries
        desc : str
            Calibration description
        irf : dict
            IRF metadata dictionary
        """
        # Set standard CIF keywords
        hdu.card('CSYSNAME', 'XMA_POL', '')
        hdu.card('CCLS0001', irf['CAL_CLASS'], 'Calibration class')
        hdu.card('CDTP0001', irf['CAL_TYPE'], 'Calibration type')
        hdu.card('CCNM0001', name, 'Calibration name')

        # Set boundary keywords
        for i in range(9):
            keyname = 'CBD%d0001' % (i+1)
            if i >= len(bounds):
                value = 'NONE'
            else:
                value = bounds[i]
            hdu.card(keyname, value, 'Calibration boundary')

        # Set validity keywords
        hdu.card('CVSD0001', irf['VAL_DATE'],
                 'Calibration validity start date (UTC)')
        hdu.card('CVST0001', irf['VAL_TIME'],
                 'Calibration validity start time (UTC)')
        hdu.card('CDEC0001', desc, 'Calibration description')

        # Return
        return


    # ROOT specific private members
    def _root2caldb(self, irf, ds):
        """
        Translate ROOT to CALDB information

        Parameters
        ----------
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Import TFile from ROOT module
        from ROOT import TFile

        # Open ROOT performance file
        tfile = TFile(self['infile'].filename().url())

        # Create effective area
        self._root2ea(tfile, irf, ds)

        # Create point spread function
        self._root2psf(tfile, irf, ds)

        # Create energy dispersion
        self._root2edisp(tfile, irf, ds)

        # Create background
        self._root2bgd(tfile, irf, ds)

        # Return
        return

    def _root2ea(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB effective area extension

        The following ROOT histograms are used:
        - EffectiveAreaEtrueNoTheta2cut_offaxis -> EFFAREA
        - EffectiveAreaEtrueNoTheta2cut (for normalization)

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Generate effective area extension')

        # Get relevant ROOT histograms
        etrue = tfile.Get('EffectiveAreaEtrueNoTheta2cut_offaxis')

        # If requested then normalize the 2D histogram on the on-axis 1D
        # histogram. This assures that the 2D histogram has the same on-axis
        # effective area dependence as the 1D histogram.
        if self['norm1d'].boolean():
            etrue_1D = tfile.Get('EffectiveAreaEtrueNoTheta2cut')
            self._renorm_onaxis(etrue, etrue_1D)

        # If requested then rebin the effective area Etrue histogram. This
        # increases the number of bins by a factor 10.
        if self['rebin'].boolean():
            etrue.RebinX(10)
            neng    = etrue.GetXaxis().GetNbins()
            noffset = etrue.GetYaxis().GetNbins()
            for ioff in range(noffset):
                for ieng in range(neng):
                    value = etrue.GetBinContent(ieng+1,ioff+1) / 10.0
                    etrue.SetBinContent(ieng+1,ioff+1,value)

        # Get effective area multiplicator. This allows renormalising the
        # effective area.
        eascale = self['eascale'].real()

        # If renormalisation has been requested then do it now
        if eascale != 1.0:
            neng    = etrue.GetXaxis().GetNbins()
            noffset = etrue.GetYaxis().GetNbins()
            for ioff in range(noffset):
                for ieng in range(neng):
                    value = etrue.GetBinContent(ieng+1,ioff+1) * eascale
                    etrue.SetBinContent(ieng+1,ioff+1,value)

        # Write boundary keywords
        self._set_cif_keywords(ds['HDU_EA'], irf['EA_NAME'],
                               irf['EA_BOUNDS'], irf['EA_DESC'], irf)

        # Optionally write energy thresholds
        if self['emin'].is_valid():
            emin = self['emin'].real()
            ds['HDU_EA'].card('LO_THRES', emin, '[TeV] Low energy threshold')
        if self['emax'].is_valid():
            emin = self['emax'].real()
            ds['HDU_EA'].card('HI_THRES', emin, '[TeV] High energy threshold')

        # Create "EFFAREA" data column
        self._make_2D(etrue, ds['HDU_EA'], 'EFFAREA', 'm**2')

        # Return
        return

    def _root2psf(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB point spread function extension

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # King profile PSF
        if self['psftype'].string() == 'King':
            self._root2psf_king(tfile, irf, ds)

        # ... otherwise use Gaussian profile PSF
        else:
            self._root2psf_gauss(tfile, irf, ds)

        # Return
        return

    def _root2psf_get_r68_and_r80(self, tfile):
        """
        Generates 68% and 80% containment radii histograms from
        AngularPSF2DEtrue_offaxis histogram

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file

        Returns
        -------
        r68, r80 : Tuple of `~ROOT.THF2`
            68% and 80% containment radius ROOT histograms
        """
        # Get relevant ROOT histograms
        psf = tfile.Get('AngularPSF2DEtrue_offaxis')

        # Extract axes
        energies = psf.GetXaxis()
        dists    = psf.GetYaxis()
        offsets  = psf.GetZaxis()

        # Extract number of bins in histogram
        neng    = energies.GetNbins()
        ndist   = dists.GetNbins()
        noffset = offsets.GetNbins()

        # Import TH2F histograms from ROOT module
        from ROOT import TH2F

        # Create r68 and r80 histograms
        r68 = TH2F('AngRes68Etrue_offaxis',
                   'Angular resolution (68% containment)',
                   neng,    energies.GetBinLowEdge(1), energies.GetBinUpEdge(neng),
                   noffset, offsets.GetBinLowEdge(1),  offsets.GetBinUpEdge(noffset))
        r80 = TH2F('AngRes80Etrue_offaxis',
                   'Angular resolution (80% containment)',
                   neng,    energies.GetBinLowEdge(1), energies.GetBinUpEdge(neng),
                   noffset, offsets.GetBinLowEdge(1),  offsets.GetBinUpEdge(noffset))

        # Loop over all energy and offset angle bins
        for ieng in range(neng):
            for ioff in range(noffset):

                # Compute PDF integral
                pdf_integral = 0.0
                for idist in range(ndist):
                    pdf_integral += psf.GetBinContent(ieng+1,idist+1,ioff+1)

                # Compute 68% containment radius
                pdf_68    = 0.0
                r68_angle = 0.0
                for idist in range(ndist):
                    pdf_68 += psf.GetBinContent(ieng+1,idist+1,ioff+1)
                    if pdf_68 > 0.68*pdf_integral:
                        r68_angle = dists.GetBinLowEdge(idist+1)
                        break

                # Compute 80% containment radius
                pdf_80    = 0.0
                r80_angle = 0.0
                for idist in range(ndist):
                    pdf_80 += psf.GetBinContent(ieng+1,idist+1,ioff+1)
                    if pdf_80 > 0.80*pdf_integral:
                        r80_angle = dists.GetBinLowEdge(idist+1)
                        break

                # Set histogram values
                r68.SetBinContent(ieng+1,ioff+1,r68_angle)
                r80.SetBinContent(ieng+1,ioff+1,r80_angle)

        # Return
        return r68, r80

    def _root2psf_gauss(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB point spread function extension

        The following ROOT histograms are used:
        - 1/(2*pi*SIGMA_1) -> SCALE
        - AngRes_offaxis -> SIGMA_1 (scaling: 1/0.8)
        - 0.0 -> AMPL_2
        - 0.0 -> SIGMA_2
        - 0.0 -> AMPL_3
        - 0.0 -> SIGMA_3

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Write header
        self._log_header1(gammalib.TERSE,
                          'Generate Gaussian point spread function extension')

        # Get relevant ROOT histograms
        r68, _ = self._root2psf_get_r68_and_r80(tfile)

        # Extract number of bins in histogram
        neng    = r68.GetXaxis().GetNbins()
        noffset = r68.GetYaxis().GetNbins()

        # Converts 68% -> 1 sigma
        r68_to_sigma = 0.6624305
        for ioff in range(noffset):
            for ieng in range(neng):
                sigma = r68.GetBinContent(ieng+1,ioff+1) * r68_to_sigma
                r68.SetBinContent(ieng+1,ioff+1,sigma)

        # Compute scale histogram
        scale = r68.Clone()
        for ioff in range(noffset):
            for ieng in range(neng):
                integral = 2.0 * math.pi * r68.GetBinContent(ieng+1,ioff+1)
                if integral > 0.0:
                    value = 1.0 / integral
                else:
                    value = 0.0
                scale.SetBinContent(ieng+1,ioff+1,value)

        # Set zero histogram
        zero = r68.Clone()
        for ioff in range(noffset):
            for ieng in range(neng):
                zero.SetBinContent(ieng+1,ioff+1,0.0)

        # Set boundaries
        irf['PSF_BOUNDS'].append('PSF(GAUSS)')

        # Write boundary keywords
        self._set_cif_keywords(ds['HDU_PSF'], irf['PSF_NAME'],
                               irf['PSF_BOUNDS'], irf['PSF_DESC'], irf)

        # Create "SCALE" data column
        self._make_2D(scale, ds['HDU_PSF'], 'SCALE', '')

        # Create "SIGMA_1" data column
        self._make_2D(r68, ds['HDU_PSF'], 'SIGMA_1', 'deg')

        # Create "AMPL_2" data column
        self._make_2D(zero, ds['HDU_PSF'], 'AMPL_2', '')

        # Create "SIGMA_2" data column
        self._make_2D(zero, ds['HDU_PSF'], 'SIGMA_2', 'deg')

        # Create "AMPL_3" data column
        self._make_2D(zero, ds['HDU_PSF'], 'AMPL_3', '')

        # Create "SIGMA_3" data column
        self._make_2D(zero, ds['HDU_PSF'], 'SIGMA_3', 'deg')

        # Return
        return

    def _root2psf_king(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB point spread function extension

        The following ROOT histograms are used:
        - AngRes_offaxis
        - AngRes80_offaxis

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Write header
        self._log_header1(gammalib.TERSE,
                          'Generate King point spread function extension')

        # Get relevant ROOT histograms
        r68, r80 = self._root2psf_get_r68_and_r80(tfile)

        # Extract number of bins in histogram
        neng    = r68.GetXaxis().GetNbins()
        noffset = r68.GetYaxis().GetNbins()

        # Initialise parameter maps by cloning the r68 2D map
        gamma2D = r68.Clone()
        sigma2D = r68.Clone()

        # Compute gamma and sigma values
        for ioff in range(noffset):

            # Initialise last results
            last_gamma = 0.0
            last_sigma = 0.0

            # Loop over all energies
            for ieng in range(neng):

                # Extract radii
                r_68 = r68.GetBinContent(ieng+1,ioff+1)
                r_80 = r80.GetBinContent(ieng+1,ioff+1)

                # Initialise results
                gamma = 0.0
                sigma = 0.0

                # Continue only if both radii are positive
                if r_68 > 0 and r_80 > 0:

                    # Derive constants for equation to solve
                    a = 1.0 - 0.68
                    b = 1.0 - 0.80
                    c = r_68*r_68/(r_80*r_80)

                    # Solve equation (a^x-1)/(b^x-1)=c for x using secant
                    # method. Stop when we are better than 1e-6.
                    x1   = -0.5
                    x2   = -1.0
                    f1   = (math.pow(a,x1) - 1.0)/(math.pow(b,x1) - 1.0) - c
                    f2   = (math.pow(a,x2) - 1.0)/(math.pow(b,x2) - 1.0) - c
                    while True:
                        x     = x1 - f1 * (x1-x2)/(f1-f2)
                        f     = (math.pow(a,x) - 1.0)/(math.pow(b,x) - 1.0) - c
                        if abs(f) < 1.0e-6:
                            break
                        else:
                            f2 = f1
                            x2 = x1
                            f1 = f
                            x1 = x

                    # Compute gamma.
                    if x < 0.0:
                        gamma = 1.0 - 1.0/x
                    else:
                        gamma = 1.0

                    # Compute sigma
                    denominator = 2.0 * gamma * (math.pow(a, x) - 1.0)
                    if denominator > 0.0:
                        sigma = r_68 * math.sqrt(1.0/denominator)
                    else:
                        denominator = 2.0 * gamma * (math.pow(b, x) - 1.0)
                        if denominator > 0.0:
                            sigma = r_80 * math.sqrt(1.0/denominator)
                        else:
                            gamma = 0.0
                            sigma = 0.0

                    # Handle special case that no results were found.
                    # This takes care of pixels that are ill defined
                    # in the MC file.
                    if gamma == 0.0 and sigma == 0.0:
                        gamma  = last_gamma
                        sigma  = last_sigma
                        status = ' (use last)'
                    else:
                        status = ''

                    # Log results
                    self._log_value(gammalib.EXPLICIT,
                                    'ieng=%d ioff=%d%s' % (ieng,ioff,status),
                                    'r68=%f r80=%f gamma=%f sigma=%f' %
                                    (r_68, r_80, gamma, sigma))

                    # Store surrent result as last result
                    last_gamma = gamma
                    last_sigma = sigma

                # Set bin contents
                gamma2D.SetBinContent(ieng+1,ioff+1,gamma)
                sigma2D.SetBinContent(ieng+1,ioff+1,sigma)

        # Set boundaries
        irf['PSF_BOUNDS'].append('PSF(KING)')

        # Write boundary keywords
        self._set_cif_keywords(ds['HDU_PSF'], irf['PSF_NAME'],
                               irf['PSF_BOUNDS'], irf['PSF_DESC'], irf)

        # Create "GAMMA" data column
        self._make_2D(gamma2D, ds['HDU_PSF'], 'GAMMA', '')

        # Create "SIGMA" data column
        self._make_2D(sigma2D, ds['HDU_PSF'], 'SIGMA', 'deg')

        # Return
        return

    def _root2edisp_get_migra(self, tfile):
        """
        Generates migration matrix histogram from AngularPSF2DEtrue_offaxis
        histogram

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file

        Returns
        -------
        migra : `~ROOT.TH3F`
            Migration matrix
        """
        # Get relevant ROOT histograms
        matrix = tfile.Get('MigMatrixNoTheta2cut_offaxis')

        # Extract axes
        ereco   = matrix.GetXaxis()
        etrue   = matrix.GetYaxis()
        offsets = matrix.GetZaxis()

        # Extract number of bins in histogram
        nereco  = ereco.GetNbins()
        netrue  = etrue.GetNbins()
        noffset = offsets.GetNbins()

        # Set number of migration bins
        nmigra = 300

        # Import TH3F histogram from ROOT module
        from ROOT import TH3F

        # Create migra histograms
        migra = TH3F('EestOverEtrue_offaxis',
                     'Migration',
                     netrue,  etrue.GetBinLowEdge(1),   etrue.GetBinUpEdge(netrue),
                     nmigra,  0.0,                      3.0,
                     noffset, offsets.GetBinLowEdge(1), offsets.GetBinUpEdge(noffset))

        # Loop over true energies
        for ietrue in range(netrue):

            # Get true energy in TeV
            e_true = math.pow(10.0, migra.GetXaxis().GetBinCenter(ietrue+1))

            # Loop over migration values
            for imigra in range(nmigra):

                # Get migration value
                migration = migra.GetYaxis().GetBinCenter(imigra+1)

                # Compute reconstructed energy in TeV
                e_reco = e_true * migration

                # Continue only if reconstructed energy is positive
                if e_reco > 0.0:

                    # Get reconstructed bin
                    ireco = ereco.FindFixBin(math.log10(e_reco))

                    # Continue only if bin is valid
                    if ireco > 0 and ireco <= nereco:

                        # Loop over offset angles
                        for ioff in range(noffset):

                            # Get bin value
                            value = matrix.GetBinContent(ireco,ietrue+1,ioff+1)

                            # Set bin value
                            migra.SetBinContent(ietrue+1,imigra+1,ioff+1,value)

        # Return
        return migra

    def _root2edisp(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB energy dispersion extension

        The following ROOT histograms are used:
        - EestOverEtrue_offaxis  -> MATRIX

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Generate energy dispersion extension')

        # Get relevant ROOT histograms
        matrix = self._root2edisp_get_migra(tfile)

        # Write boundary keywords
        self._set_cif_keywords(ds['HDU_EDISP'], irf['EDISP_NAME'],
                               irf['EDISP_BOUNDS'], irf['EDISP_DESC'], irf)

        # Create "MATRIX" data column
        self._make_3D_migra(matrix, ds['HDU_EDISP'], 'MATRIX', '')

        # Return
        return

    def _root2bgd(self, tfile, irf, ds):
        """
        Translate ROOT to CALDB background extension.

        The following ROOT histograms are used:
        - BGRatePerSqDeg_offaxis -> BKG

        Parameters
        ----------
        tfile : `~ROOT.TFile`
            ROOT file
        irf : dict
            IRF metadata dictionary
        ds : dict
            Directory structure dictionary
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Generate 3D background extension')

        # Get relevant ROOT histograms
        array = tfile.Get('BGRatePerSqDeg_offaxis')

        # Replace 2D histogram values by power law extrapolation
        self._plaw_replace(array, self['bgdethres'].real())

        # If requested then fill-in empty values in 2D histogram
        if self['bgdinfill'].boolean():
            self._infill_bgd(array)

        # If requested then normalize the 2D histogram on the on-axis 1D
        # histogram. This assures that the 2D histogram has the same on-axis
        # effective area dependence as the 1D histogram.
        if self['norm1d'].boolean():
            array_1D = tfile.Get('BGRatePerSqDeg')
            self._renorm_onaxis(array, array_1D)

        # Write boundary keywords
        self._set_cif_keywords(ds['HDU_BKG'], irf['BKG_NAME'],
                               irf['BKG_BOUNDS'], irf['BKG_DESC'], irf)

        # Create "BKG" data column
        self._make_3D(array, ds['HDU_BKG'], 'BKG', '1/(MeV s sr)')

        # Return
        return

    def _append_column_axis(self, hdu, name, unit, axis, log=False):
        """
        Append column of ROOT axis values to HDU

        Parameters
        ----------
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Column name
        unit : str
            Column unit
        axis : `~ROOT.TAxis`
            ROOT histogram axis
        log : bool, optional
            Axis is logarithmic
        """
        # Continue only if columns does not yet exist
        if not hdu.contains(name):

            # Write header
            self._log_header3(gammalib.TERSE, 'Append axis column "%s"' % name)

            # Get number of axis bins
            nbins = axis.GetNbins()

            # Append column and set column unit
            hdu.append(gammalib.GFitsTableFloatCol(name, 1, nbins))
            hdu[name].unit(unit)

            # Do we have a LO axis? If not we have a HI axis
            if name[-2:] == 'LO':
                low = True
            else:
                low = False

            # Fill column
            for i in range(nbins):
                if low:
                    value = axis.GetBinLowEdge(i+1)
                else:
                    value = axis.GetBinUpEdge(i+1)
                if log:
                    value = pow(10.0, value)
                hdu[name][0,i] = value

            # Log values
            self._log_value(gammalib.NORMAL, 'Number of axis bins', nbins)
            self._log_value(gammalib.NORMAL, 'Unit', unit)

        # Return
        return

    def _append_column_values(self, hdu, name, unit, values):
        """
        Append column of values to HDU

        Parameters
        ----------
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Column name
        unit : str
            Column unit
        values : list of float
            Axis values
        """
        # Continue only if columns does not yet exist
        if not hdu.contains(name):

            # Write header
            self._log_header3(gammalib.TERSE, 'Append value column "%s"' % name)

            # Get number of values
            nbins = len(values)

            # Append column and set column unit
            hdu.append(gammalib.GFitsTableFloatCol(name, 1, nbins))
            hdu[name].unit(unit)

            # Fill column
            for i in range(nbins):
                hdu[name][0,i] = values[i]

            # Log values
            self._log_value(gammalib.NORMAL, 'Number of values', nbins)
            self._log_value(gammalib.NORMAL, 'Unit', unit)

        # Return
        return

    def _renorm_onaxis(self, hist2D, hist1D):
        """
        Renormalise 2D histogram (energy,offset) on 1D histogram (energy)

        This method makes sure that a 2D histogram has the same onaxis values
        as the corresponding 1D histogram.

        Parameters
        ----------
        hist2D : `~ROOT.TH2F`
            ROOT 2D histogram
        hist1D : `~ROOT.TH1F`
            ROOT 1D histogram
        """
        # Get 2D dimensions
        neng    = hist2D.GetXaxis().GetNbins()
        noffset = hist2D.GetYaxis().GetNbins()

        # Continue only if the 1D and 2D histograms have the same number
        # of energy bins
        if neng != hist1D.GetXaxis().GetNbins():

            # Log if energy binning differs
            self._log_value(gammalib.TERSE, 'On-axis normalisation',
                            'Impossible since energy binning of 1D histogram '
                            '"%s" does not match that of 2D histogram "%s".' %
                            (hist1D.GetName(), hist2D.GetName()))
        else:
            # Log on-axis normalisation
            self._log_value(gammalib.TERSE, 'On-axis normalisation',
                            'On-axis values of 1D histogram "%s" imposed on '
                            '2D histogram "%s".' %
                            (hist1D.GetName(), hist2D.GetName()))

            # Get a copy of 2D histogram
            hist2D_copy = hist2D.Clone()

            # Divide 2D histogram by onaxis values to remove the energy
            # dependence
            for ieng in range(neng):
                onaxis = hist1D.GetBinContent(ieng+1)
                if onaxis > 0.0:
                    for ioff in range(noffset):
                        value = hist2D.GetBinContent(ieng+1,ioff+1) / onaxis
                        hist2D.SetBinContent(ieng+1,ioff+1,value)

            # Smooth in energy direction to reduce statistical fluctuations
            for ieng in range(neng):
                for ioff in range(noffset):
                    if ieng == 0:
                        value = (2.0 * hist2D.GetBinContent(ieng+1,ioff+1) +
                                 1.0 * hist2D.GetBinContent(ieng+2,ioff+1)) / 3.0
                    elif ieng == neng-1:
                        value = (2.0 * hist2D.GetBinContent(ieng+1,ioff+1) +
                                 1.0 * hist2D.GetBinContent(ieng+0,ioff+1)) / 3.0
                    else:
                        value = (2.0 * hist2D.GetBinContent(ieng+1,ioff+1) +
                                 1.0 * hist2D.GetBinContent(ieng+0,ioff+1) +
                                 1.0 * hist2D.GetBinContent(ieng+2,ioff+1)) / 4.0
                    hist2D_copy.SetBinContent(ieng+1,ioff+1,value)

            # Normalize now for an onaxis value of 1
            for ieng in range(neng):
                onaxis = hist2D_copy.GetBinContent(ieng+1,1)
                if onaxis > 0.0:
                    for ioff in range(noffset):
                        value = hist2D_copy.GetBinContent(ieng+1,ioff+1) / onaxis
                        hist2D_copy.SetBinContent(ieng+1,ioff+1,value)

            # Put back the energy dependence
            for ieng in range(neng):
                onaxis = hist1D.GetBinContent(ieng+1)
                for ioff in range(noffset):
                    value = hist2D_copy.GetBinContent(ieng+1,ioff+1) * onaxis
                    hist2D.SetBinContent(ieng+1,ioff+1,value)

        # Return
        return

    def _plaw_replace(self, hist2D, ethres):
        """
        Replace background rate values by power law values

        Parameters
        ----------
        hist2D : `~ROOT.TH2F`
            ROOT 2D histogram
        ethres : float
            Energy threshold
        """
        # Extract energy and offset angle vectors
        energies = hist2D.GetXaxis()
        neng     = energies.GetNbins()
        offsets  = hist2D.GetYaxis()
        noffset  = offsets.GetNbins()

        # Continue only if threshold is low enough
        if ethres < math.pow(10.0, energies.GetBinUpEdge(neng)):

            # Determine mean energy values
            engs = []
            for ieng in range(neng):
                energy = 0.5 * (math.pow(10.0, energies.GetBinLowEdge(ieng+1)) +
                                math.pow(10.0, energies.GetBinUpEdge(ieng+1)))
                engs.append(energy)

            # Loop over all offaxis angles
            for ioff in range(noffset):

                # Determine offset angle (for logging)
                offset = 0.5 * (offsets.GetBinLowEdge(ioff+1) +
                                offsets.GetBinUpEdge(ioff+1))

                # Initialise vectors
                axis  = []
                rates = []

                # Extract values
                for ieng in range(neng):
                    energy = engs[ieng]
                    rate   = hist2D.GetBinContent(ieng+1,ioff+1)
                    axis.append(energy)
                    rates.append(rate)

                # Determine minimum energy (ignoring the maximum energy as
                # we do not need it)
                emin = self._plaw_energy_range(axis, rates)[0]

                # Get power law coefficients
                coeff = self._plaw_coefficients(axis, rates, emin, ethres)

                # Replace histogram values by power law values if coefficients
                # are valid
                if coeff['m'] != 0.0 and coeff['t'] != 0.0:
                    for ieng in range(neng):
                        energy = engs[ieng]
                        if energy > ethres:
                            value = self._plaw_value(coeff, energy)
                            hist2D.SetBinContent(ieng+1,ioff+1,value)

                            # Log power law replacement
                            self._log_value(gammalib.NORMAL,
                                            'Power law replacement',
                                            '%e cts/(s deg**2) (E=%8.4f TeV (%d) '
                                            'Off=%.2f deg (%d)' %
                                            (value,engs[ieng],ieng,offset,ioff))

        # Return
        return

    def _plaw_energy_range(self, energies, rates):
        """
        Determine the energy range over which the power law is valid

        Parameters
        ----------
        energies : list of float
            Energy values (TeV)
        rates : list of float
            Background rate values
        """
        # Find energy at which background rate takes a maximum
        min_energy = 0.0
        max_rate   = 0.0
        for ieng, energy in enumerate(energies):
            if rates[ieng] > max_rate:
                max_rate   = rates[ieng]
                min_energy = energy

        # Find last data point
        max_energy = 0.0
        for ieng in range(len(energies)-1,-1,-1):
            if rates[ieng] > 0.0:
                max_energy = energies[ieng]
                break

        # Return
        return (min_energy, max_energy)

    def _plaw_coefficients(self, energies, rates, emin, emax):
        """
        Determines the power law coefficients

        Parameters
        ----------
        energies : list of float
            Energy values (TeV)
        rates : list of float
            Background rate values
        emin : float
            Minimum energy (TeV)
        emax : float
            Maximum energy (TeV)
        """
        # Determine nearest energy indices
        i_emin     = 0
        i_emax     = 0
        delta_emin = 1.0e30
        delta_emax = 1.0e30
        for ieng, energy in enumerate(energies):
            if abs(energy-emin) < delta_emin:
                delta_emin = abs(energy-emin)
                i_emin     = ieng
            if abs(energy-emax) < delta_emax:
                delta_emax = abs(energy-emax)
                i_emax     = ieng

        # Determine coefficients of power law
        while rates[i_emax] == 0.0 and i_emax > i_emin:
            i_emax -= 1
        if rates[i_emax] > 0.0:
            x1 = math.log10(energies[i_emin])
            x2 = math.log10(energies[i_emax])
            y1 = math.log10(rates[i_emin])
            y2 = math.log10(rates[i_emax])
            m  = (y2-y1)/(x2-x1)
            t  = y1 - m * x1
        else:
            m = 0.0
            t = 0.0

        # Return coefficients
        return {'m': m, 't': t}

    def _plaw_value(self, coeff, energy):
        """
        Determine the power law value

        Parameters
        ----------
        coeff : dict
            Dictionary of coefficients
        energy : float
            Energy value (TeV)
        """
        # Compute value
        value = math.pow(10.0, coeff['m'] * math.log10(energy) + coeff['t'])

        # Return value
        return value

    def _infill_bgd(self, hist2D):
        """
        Fill all empty bins in 2D histogram by preceeding values in energy

        Parameters
        ----------
        hist2D : `~ROOT.TH2F`
            ROOT 2D histogram
        """
        # Extract energy and offset angle vectors
        energies = hist2D.GetXaxis()
        neng     = energies.GetNbins()
        offsets  = hist2D.GetYaxis()
        noffset  = offsets.GetNbins()

        # Determine mean energy values (for logging)
        engs = []
        for ieng in range(neng):
            energy = 0.5 * (math.pow(10.0, energies.GetBinLowEdge(ieng+1)) +
                            math.pow(10.0, energies.GetBinUpEdge(ieng+1)))
            engs.append(energy)

        # Loop over all offaxis angles
        for ioff in range(noffset):

            # Determine offset angle (for logging)
            offset = 0.5 * (offsets.GetBinLowEdge(ioff+1) +
                            offsets.GetBinUpEdge(ioff+1))

            # Initialise indices and value
            i_start = -1
            i_stop  = -1
            value   = 0.0

            # First start and stop indices for infill
            for ieng in range(neng):
                if hist2D.GetBinContent(ieng+1,ioff+1) > 0.0:
                    i_start = ieng
                    value   = hist2D.GetBinContent(ieng+1,ioff+1)
                    break
            for ieng in range(neng-1, -1, -1):
                if hist2D.GetBinContent(ieng+1,ioff+1) > 0.0:
                    i_stop = ieng
                    break

            # If indices are valid then fill in background rates
            if i_start > -1 and i_stop > -1:
                for ieng in range(i_start, i_stop+1):
                    if hist2D.GetBinContent(ieng+1,ioff+1) == 0.0:
                        hist2D.SetBinContent(ieng+1,ioff+1,value)

                        # Log background infill
                        self._log_value(gammalib.NORMAL,
                                        'Background infill',
                                        '%e cts/(s deg**2) (E=%8.4f TeV (%d) '
                                        'Off=%.2f deg (%d)' %
                                        (value,engs[ieng],ieng,offset,ioff))

                    else:
                        value = hist2D.GetBinContent(ieng+1,ioff+1)

        # Return
        return

    def _make_2D(self, array, hdu, name, unit, scale=1.0):
        """
        Make 2D data column as function of energy and offset angle

        If the HDU has already the energy and offset angle columns, this method
        will simply add another data column. If name==None, the method will not
        append any data column.

        Parameters
        ----------
        array : `~ROOT.TH2F`
            ROOT 2D histogram
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Data column name
        unit : str
            Data unit
        scale : float, optional
            Scaling factor for histogram values
        """
        # Write header
        self._log_header3(gammalib.TERSE, 'Make 2D data column "%s"' % name)

        # Extract energy and offset angle vectors
        energies = array.GetXaxis()
        offsets  = array.GetYaxis()
        neng     = energies.GetNbins()
        noffset  = offsets.GetNbins()

        # Log parameters
        self._log_value(gammalib.NORMAL, 'Number of energies', neng)
        self._log_value(gammalib.NORMAL, 'Number of offsets', noffset)

        # Append axis columns to HDU
        self._append_column_axis(hdu, 'ENERG_LO', 'TeV', energies, log=True)
        self._append_column_axis(hdu, 'ENERG_HI', 'TeV', energies, log=True)
        self._append_column_axis(hdu, 'THETA_LO', 'deg', offsets)
        self._append_column_axis(hdu, 'THETA_HI', 'deg', offsets)

        # Append array column
        if name != None and not hdu.contains(name):
            hdu.append(gammalib.GFitsTableFloatCol(name, 1, neng*noffset))
            hdu[name].unit(unit)
            hdu[name].dim([neng, noffset])
            for ioff in range(noffset):
                for ieng in range(neng):
                    index = ieng + ioff * neng
                    value = array.GetBinContent(ieng+1,ioff+1)
                    hdu[name][0,index] = value * scale

        # Collect boundary information
        bd_eng = 'ENERG(%.4f-%.2f)TeV' % \
                 (pow(10.0, energies.GetBinLowEdge(1)), \
                  pow(10.0, energies.GetBinUpEdge(neng)))
        bd_off = 'THETA(%.2f-%.2f)deg' % \
                 (offsets.GetBinLowEdge(1), \
                  offsets.GetBinUpEdge(noffset))
        bd_phi = 'PHI(0-360)deg'
        bounds = [bd_eng, bd_off, bd_phi]

        # Return boundary information
        return bounds

    def _make_3D(self, array, hdu, name, unit):
        """
        Make 3D data column as function of DETX, DETY and energy

        If the HDU has already the energy and offset angle columns, this method
        will simply add another data column. If name==None, the method will not
        append any data column.

        Parameters
        ----------
        array : `~ROOT.TH2F`
            ROOT 2D histogram
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Data column name
        unit : str
            Data unit
        """
        # Write header
        self._log_header3(gammalib.TERSE, 'Make 3D data column "%s"' % name)

        # Get User parameters
        scale      = self['bgdscale'].real()
        oversample = self['bgdoversample'].integer()

        # Set constants
        deg2sr = 0.01745329*0.01745329

        # Extract energy and offset angle vectors
        energies  = array.GetXaxis()
        neng      = energies.GetNbins()
        offsets   = array.GetYaxis()
        noffset   = offsets.GetNbins()
        theta_max = offsets.GetBinUpEdge(noffset) # Maximum theta
        ewidth    = [] # in MeV
        for ieng in range(neng):
            ewidth.append(pow(10.0, energies.GetBinUpEdge(ieng+1) +6.0) -
                          pow(10.0, energies.GetBinLowEdge(ieng+1)+6.0))

        # Log parameters
        self._log_value(gammalib.NORMAL, 'Scale', scale)
        self._log_value(gammalib.NORMAL, 'Oversample', oversample)
        self._log_value(gammalib.NORMAL, 'Number of energies', neng)
        self._log_value(gammalib.NORMAL, 'Number of offsets', noffset)
        self._log_value(gammalib.NORMAL, 'Maximum offset', theta_max)

        # Build DETX and DETY axes
        ndet     = array.GetYaxis().GetNbins()
        ndets    = 2*ndet*oversample
        det_max  = array.GetYaxis().GetBinUpEdge(ndet)
        det_bin  = det_max/float(ndet*oversample)
        dets_lo  = []
        dets_hi  = []
        dets2    = []
        for i in range(ndets):
            det_lo  = -det_max + i*det_bin
            det_hi  = det_lo + det_bin
            det_val = 0.5*(det_lo + det_hi)
            dets_lo.append(det_lo)
            dets_hi.append(det_hi)
            dets2.append(det_val*det_val)

        # Log parameters
        self._log_value(gammalib.NORMAL, 'Number of DETXY bins', ndets)
        self._log_value(gammalib.NORMAL, 'Maximum DETXY', det_max)
        self._log_value(gammalib.NORMAL, 'DETXY binsize', det_bin)

        # Append DETX_LO, DETX_HI, DETY_LO and DETY_HI columns
        self._append_column_values(hdu, 'DETX_LO', 'deg', dets_lo)
        self._append_column_values(hdu, 'DETX_HI', 'deg', dets_hi)
        self._append_column_values(hdu, 'DETY_LO', 'deg', dets_lo)
        self._append_column_values(hdu, 'DETY_HI', 'deg', dets_hi)

        # Append ENERG_LO and ENERG_HI columns
        self._append_column_axis(hdu, 'ENERG_LO', 'TeV', energies, log=True)
        self._append_column_axis(hdu, 'ENERG_HI', 'TeV', energies, log=True)

        # Append array column
        if name != None and not hdu.contains(name):
            hdu.append(gammalib.GFitsTableFloatCol(name, 1, ndets*ndets*neng))
            hdu[name].unit(unit)
            hdu[name].dim([ndets,ndets,neng])
            for ix in range(ndets):
                for iy in range(ndets):
                    for ieng in range(neng):
                        index = ix + (iy + ieng * ndets) * ndets
                        theta = math.sqrt(dets2[ix] + dets2[iy])
                        if theta < theta_max:
                            binsq = array.Interpolate(energies.GetBinCenter(ieng+1),
                                                      theta)
                            value = binsq / deg2sr / (ewidth[ieng])
                            hdu[name][0,index] = value * scale

        # Collect boundary information
        bd_detx = 'DETX(%.2f-%.2f)deg' % (-det_max, det_max)
        bd_dety = 'DETY(%.2f-%.2f)deg' % (-det_max, det_max)
        bd_eng  = 'ENERG(%.4f-%.2f)TeV' % \
                  (pow(10.0, energies.GetBinLowEdge(1)),
                   pow(10.0, energies.GetBinUpEdge(neng)))
        bounds  = [bd_detx, bd_dety, bd_eng]

        # Return boundary information
        return bounds

    def _make_3D_migra(self, array, hdu, name, unit, scale=1.0):
        """
        Make 3D data column as function of ENERG, MIGRA and THETA

        If the HDU has already the energy and offset angle columns, this method
        will simply add another data column. If name==None, the method will not
        append any data column.

        Parameters
        ----------
        array : `~ROOT.TH3F`
            ROOT 2D histogram
        hdu : `~gammalib.GFitsHDU`
            FITS HDU
        name : str
            Data column name
        unit : str
            Data unit
        scale : float, optional
            Scaling factor for histogram values
        """
        # Write header
        self._log_header3(gammalib.TERSE,
                          'Make 3D migration matrix data column "%s"' % name)

        # Extract Etrue, Eobs/Etrue and offset angle vectors
        etrue   = array.GetXaxis()
        netrue  = etrue.GetNbins()
        migra   = array.GetYaxis()
        nmigra  = migra.GetNbins()
        offsets = array.GetZaxis()
        noffset = offsets.GetNbins()
        ewidth  = [] # in MeV
        for ieng in range(netrue):
            ewidth.append(pow(10.0, etrue.GetBinUpEdge(ieng+1) +6.0) - \
                          pow(10.0, etrue.GetBinLowEdge(ieng+1)+6.0))

        # Log parameters
        self._log_value(gammalib.NORMAL, 'Number of energies', netrue)
        self._log_value(gammalib.NORMAL, 'Number of migrations', nmigra)
        self._log_value(gammalib.NORMAL, 'Number of offsets', noffset)

        # Append axis columns to HDU
        self._append_column_axis(hdu, 'ENERG_LO', 'TeV', etrue, log=True)
        self._append_column_axis(hdu, 'ENERG_HI', 'TeV', etrue, log=True)
        self._append_column_axis(hdu, 'MIGRA_LO', '', migra)
        self._append_column_axis(hdu, 'MIGRA_HI', '', migra)
        self._append_column_axis(hdu, 'THETA_LO', 'deg', offsets)
        self._append_column_axis(hdu, 'THETA_HI', 'deg', offsets)

        # Append migration matrix to HDU
        if name != None and not hdu.contains(name):
            hdu.append(gammalib.GFitsTableFloatCol(name, 1, netrue*nmigra*noffset))
            hdu[name].unit(unit)
            hdu[name].dim([netrue, nmigra, noffset])
            for ioff in range(noffset):
                for imigra in range(nmigra):
                    for ieng in range(netrue):
                        index = ieng + (imigra + ioff * nmigra) * netrue
                        value = array.GetBinContent(ieng+1,imigra+1,ioff+1)
                        hdu[name][0,index] = value * scale

        # Collect boundary information
        bd_eng   = 'ENERG(%.4f-%.2f)TeV' % \
                   (pow(10.0, etrue.GetBinLowEdge(1)),
                    pow(10.0, etrue.GetBinUpEdge(netrue)))
        bd_migra = 'MIGRA(%.3f-%.3f)' % \
                   (migra.GetBinLowEdge(1),
                    migra.GetBinUpEdge(nmigra))
        bd_off   = 'THETA(%.2f-%.2f)deg' % \
                   (offsets.GetBinLowEdge(1),
                    offsets.GetBinUpEdge(noffset))
        bd_phi   = 'PHI(0-360)deg'
        bounds   = [bd_eng, bd_migra, bd_off, bd_phi]

        # Return boundary information
        return bounds


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Optional ROOT import
        try:
            from ROOT import TFile, TH2F, TH3F
            _has_root = True
        except ImportError:
            _has_root = False

        # Warn if ROOT module is missing
        if not _has_root:
            gammalib.warning('csroot2caldb', 'ROOT Python module not present, '
                             'script will create empty IRF.')

        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Initialise IRF metadata
        irf = self._init_metadata()

        # Create directory structure
        ds = self._make_dirs('file', irf)

        # Open calibration files
        self._open(irf, ds)

        # Translate ROOT to CALDB information
        if _has_root:
            self._root2caldb(irf, ds)

        # Close calibration files
        self._close(irf, ds)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csroot2caldb(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Generates pointing patterns for CTA observations
#
# Copyright (C) 2015-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import math
import gammalib


# ============================================================ #
# Get positions of pointings on sky for an interleaved pattern #
# ============================================================ #
def get_positions(xmin, xmax, ymin, ymax, step):
    """
    Get pointing positions for an interleaved pattern

    Get positions for a patch of interleaved pointings on the sky.
    The function does respect the latitude dependence of the
    pointing distance. The step parameter is for a latitude of
    zero.

    Parameters
    ----------
    xmin : float
        Minimum longitude (deg)
    xmax : float
        Maximum longitude (deg)
    ymin : float
        Minimum latitude (deg)
    ymax : float
        Maximum latitude (deg)
    step : float
        Pointing step size (deg)

    Returns
    -------
    positions : list of dict
        List of dictionaries with x and y positions
    """
    # Initialise positions
    positions = []

    # Determine ystep and number of pointing rows
    ystep   = step * math.sqrt(0.75)
    ny      = int((ymax-ymin)/ystep+1.5)
    rescale = (ymax-ymin)/(ystep*(ny-1))
    ystep  *= rescale
    print('Number of rows ....: %d' % (ny))

    # Loop over rows
    y = ymin
    for row in range(ny):

        # Determine xstep and number of pointings in row. The xstep increases
        # with the cosine of the latitude so that the distance of the step
        # is invariant of latitude.
        xstep   = step / math.cos(gammalib.deg2rad * y)
        nx      = int((xmax-xmin)/xstep+1.5)
        xstep   = (xmax-xmin)/float(nx-1)

        # Set x offset. For every second row the x position is displace by
        # half a step size.
        if row % 2 == 0:
            x = xmin
        else:
            x = xmin + 0.5 * xstep

        # Append pointings
        for pnt in range(nx):
            positions.append({'x': x, 'y': y})
            x += xstep
            if x >= 360.0:
                x -= 360.0

        # Increment y
        y += ystep

    # Return positions
    return positions


# ============== #
# Set start time #
# ============== #
def set_tmin_for_next_pointing(tmin, duration):
    """
    Set start time for next pointing

    Adds duration and 2 min for slew, and takes care of day and night. It is
    assumed that the night lasts from [0,10] and the day lasts from [10,24]

    Parameters
    ----------
    tmin : float
        Start time of current pointing (sec)
    duration : float
        Duration of current pointing (sec)

    Returns
    -------
    tmin : float
        Start time of new pointing
    """
    # Update start time, add 2 min for slew
    tmin += duration + 120.0

    # If we have more than 10 hours in day then go to next day
    hours_in_day = tmin/3600.0 % 24.0
    if hours_in_day > 10.0:
        days  = int(tmin/86400.0)
        days += 1
        tmin  = float(days) * 86400.0

    # Dump result
    #print(tmin, hours_in_day)

    # Return start time
    return tmin


# ====================== #
# Setup patch on the sky #
# ====================== #
def set_patch(tmin, lmin=-30.0, lmax=30.0, bmin=-1.5, bmax=+1.5,
              separation=3.0, hours=100.0, run_duration=30.0,
              site=None, caldb='prod2', lst=True, autodec=0.0, north=True):
    """
    Setup pointing patch on the sky

    Parameters
    ----------
    tmin : float
        Observation start time (seconds)
    lmin : float, optional
        Minimum Galactic Longitude (deg)
    lmax : float, optional
        Maximum Galactic Longitude (deg)
    bmin : float, optional
        Minimum Galactic Latitude (deg)
    bmax : float, optional
        Maximum Galactic Latitude (deg)
    separation : float, optional
        Pointing separation (deg)
    hours : float, optional
        Total observation duration (hours)
    run_duration : float, optional
        Run duration (min)
    site : str, optional
        Array site (one of 'Automatic', 'South', 'North')
    lst : bool, optional
        Use LSTs
    autodec : float, optional
        Declination for switching automatically between South and North site
    north : bool, optional
        Also use North array

    Returns
    -------
    obsdef : list of dict
        List of pointing definitions
    """
    # Initialise observation definition
    obsdef = []

    # Get pointing positions
    pointings = get_positions(lmin, lmax, bmin, bmax, separation)

    # Determine number of pointings and pointing duration
    n_pnt    = len(pointings)
    duration = hours*3600.0/float(n_pnt)

    # Observing time collectors
    exposure_south = 0.0
    exposure_north = 0.0

    # Set observations
    for pnt in pointings:

        # Compute number of runs for this pointing
        if run_duration > 0.0:
            n_runs   = int(duration / (run_duration*60.0) + 0.5)
            exposure = run_duration*60.0
            if n_runs < 1:
                n_runs = 1
        else:
            n_runs   = 1
            exposure = duration

        # Loop over number of runs
        for run in range(n_runs):

            # Set positions, start time and duration
            obs = {'lon': pnt['x'], 'lat': pnt['y'], \
                   'tmin': tmin, 'duration': exposure}

            # Update start time for next pointing
            tmin = set_tmin_for_next_pointing(tmin, exposure)

            # Optionally add-in site dependent IRF
            if site != None:

                # If automatic site switching is requested then set the North
                # IRF for declinations greater or equal than "autodec" and the
                # South IRF for declination smaller than "autodec".
                if site == 'Automatic':
                    pos = gammalib.GSkyDir()
                    pos.lb_deg(pnt['x'], pnt['y'])
                    dec = pos.dec_deg()
                    if (dec >= autodec):
                        obs_site = 'North'
                    else:
                        obs_site = 'South'

                # ... otherwise use the specified site, North or South
                else:
                    obs_site = site

                # Fall through if array is North array and North array should
                # not be used
                if not north and obs_site == 'North':
                    continue

                # Set IRF
                irf = set_irf(obs_site, obs, caldb, lst=lst)

                # Set IRF information
                obs['caldb'] = caldb
                obs['irf']   = irf

                # Add up exposure for a given site
                if obs_site == 'North':
                    exposure_north += exposure
                else:
                    exposure_south += exposure

            # Append observation
            obsdef.append(obs)

    # Dump statistics
    print('Number of pointings: %d (%.2f s)' % (n_pnt,duration))
    print('South array .......: %.2f hours (%.2f%%)' %
          (exposure_south/3600.0,
           100.0*exposure_south/(exposure_south+exposure_north)))
    print('North array .......: %.2f hours (%.2f%%)' %
          (exposure_north/3600.0,
           100.0*exposure_north/(exposure_south+exposure_north)))

    # Return observation definition
    return obsdef


# ================================== #
# Setup Instrument Response Function #
# ================================== #
def set_irf(site, obs, caldb, lst=True):
    """
    Setup Instrument Response Function

    Parameters
    ----------
    site : str
        Array site (either 'South' or 'North')
    obs : dict
        Dictionary with pointing information
    caldb : str
        Calibration database
    lst : bool, optional
        Use LSTs

    Returns
    -------
    irf : str
        IRF name
    """
    # Handle 'prod2'
    if caldb == 'prod2':
        if site == 'North':
            irf = 'North_50h'
        else:
            irf = 'South_50h'

    # Handle 'prod3' and 'prod3b'
    else:

        # Compute Right Ascension and Declination of pointing
        pnt = gammalib.GSkyDir()
        pnt.lb_deg(obs['lon'], obs['lat'])
        #ra  = pnt.ra_deg()
        dec = pnt.dec_deg()

        # Set geographic latitude of array
        if site == 'North':
            geolat = +28.7569
        else:
            geolat = -24.58

        # Compute best possible zenith angle
        zenith = abs(dec - geolat)

        # Set IRF
        if site == 'North':
            irf = 'North'
        else:
            irf = 'South'
        if zenith < 30.0:
            irf += '_z20'
        else:
            irf += '_z40'
        irf += '_50h'

    # Return IRF
    return irf


# =========================== #
# Setup Galactic plane survey #
# =========================== #
def set_gps(separation=3.0, bmin=-1.3, bmax=1.3, caldb='prod2', lst=True,
            north=True):
    """
    Setup Galactic plane survey

    Parameters
    ----------
    separation : float, optional
        Pointing separation (deg)
    bmin : float, optional
        Minimum Galactic Latitude (deg)
    bmax : float, optional
        Maximum Galactic Latitude (deg)
    caldb : str, optional
        Calibration database name
    lst : bool, optional
        Use LSTs
    north : bool, optional
        Also use North array

    Returns
    -------
    obsdef : list of dict
        List of pointing definitions
    """
    # Initialise observation definition
    obsdef = []

    # Initialise start time (1-1-2021)
    tmin = 7671.0 * 86400.0

    # Add inner region South
    obsdef.extend(set_patch(tmin, lmin=-60.0, lmax=60.0, bmin=bmin, bmax=bmax,
                            separation=separation, hours=780,
                            site='South', caldb=caldb, lst=lst))

    # Update start time
    tmin = set_tmin_for_next_pointing(obsdef[-1]['tmin'], obsdef[-1]['duration'])

    # Add Vela & Carina region
    obsdef.extend(set_patch(tmin, lmin=240.0, lmax=300.0, bmin=bmin, bmax=bmax,
                            separation=separation, hours=180,
                            site='South', caldb=caldb, lst=lst))

    # Update start time
    tmin = set_tmin_for_next_pointing(obsdef[-1]['tmin'], obsdef[-1]['duration'])

    # Add 210-240 region
    obsdef.extend(set_patch(tmin, lmin=210.0, lmax=240.0, bmin=bmin, bmax=bmax,
                            separation=separation, hours=60,
                            site='South', caldb=caldb, lst=lst))

    # Also use North array
    if north:

        # Initialise start time
        tmin = 7671.0 * 86400.0

        # Add Cygnus, Perseus
        obsdef.extend(set_patch(tmin, lmin=60.0, lmax=150.0,
                                bmin=bmin, bmax=bmax, separation=separation,
                                hours=450, site='North', caldb=caldb, lst=lst))

        # Update start time
        tmin = set_tmin_for_next_pointing(obsdef[-1]['tmin'],
                                          obsdef[-1]['duration'])

        # Add Anticentre
        obsdef.extend(set_patch(tmin, lmin=150.0, lmax=210.0,
                                bmin=bmin, bmax=bmax, separation=separation,
                                hours=150, site='North', caldb=caldb, lst=lst))

    # Return observation definition
    return obsdef


# ========================== #
# Setup Extragalactic survey #
# ========================== #
def set_egal(separation=3.0, caldb='prod2', lst=True, north=True):
    """
    Setup Extragalactic survey.

    Parameters
    ----------
    separation : float, optional
        Pointing separation (deg)
    caldb : str, optional
        Calibration database name
    lst : bool, optional
        Use LSTs
    north : bool, optional
        Also use North array

    Returns
    -------
    obsdef : list of dict
        List of pointing definitions
    """
    # Initialise observation definition
    obsdef = []

    # Initialise start time (1-1-2021)
    tmin = 7671.0 * 86400.0

    # Set patch
    obsdef.extend(set_patch(tmin, lmin=-90.0, lmax=90.0, bmin=+5.0, bmax=+88.0,
                            separation=separation, hours=500, run_duration=25,
                            site='Automatic', caldb=caldb, lst=lst,
                            autodec=-10.0, north=north))

    # Return observation definition
    return obsdef


# ========================= #
# Setup Galactic centre KSP #
# ========================= #
def set_gc(caldb='prod2', lst=True):
    """
    Setup Galactic centre KSP.

    Parameters
    ----------
    caldb : str, optional
        Calibration database name
    lst : bool, optional
        Use LSTs

    Returns
    -------
    obsdef : list of dict
        List of pointing definitions
    """
    # Initialise observation definition
    obsdef = []

    # Initialise start time (1-1-2021)
    tmin = 7671.0 * 86400.0

    # Central wobble
    obsdef.extend(set_patch(tmin, lmin=-1.0, lmax=1.0, bmin=-1.0, bmax=1.0,
                            separation=0.1, hours=525,
                            site='South', caldb=caldb, lst=lst))

    # Update start time
    tmin = set_tmin_for_next_pointing(obsdef[-1]['tmin'], obsdef[-1]['duration'])

    # Extended region
    obsdef.extend(set_patch(tmin, lmin=-3.0, lmax=3.0, bmin=3.0, bmax=10.0,
                            separation=0.5, hours=300,
                            site='South', caldb=caldb, lst=lst))

    # Return observation definition
    return obsdef


# ============= #
# Setup LMC KSP #
# ============= #
def set_lmc(hours=250.0, caldb='prod2', lst=True):
    """
    Setup LMC KSP.

    Parameters
    ----------
    hours : float, optional
        Total observation duration (h)
    caldb : str, optional
        Calibration database name
    lst : bool, optional
        Use LSTs

    Returns
    -------
    obsdef : list of dict
        List of pointing definitions
    """
    # Initialise observation definition
    obsdef = []

    # Initialise start time (1-1-2021)
    tmin = 7671.0 * 86400.0

    # Set LMC centre
    centre = gammalib.GSkyDir()
    centre.radec_deg(80.0, -69.0)

    # Set offset angle and number of pointings
    offset = 1.0 # degrees
    n_pnt  = 12

    # Prepare computations
    dphi     = 360.0/n_pnt
    duration = hours*3600.0/float(n_pnt)

    # Loop over pointings
    for ipnt in range(n_pnt):

        # Compute pointing direction
        pnt = centre.copy()
        pnt.rotate_deg(ipnt*dphi, offset)
        lon = pnt.l_deg()
        lat = pnt.b_deg()

        # Set positions and duration
        obs = {'lon': lon, 'lat': lat, 'tmin': tmin, 'duration': duration}

        # Set IRF
        irf = set_irf('South', obs, caldb, lst=lst)

        # Add IRF
        obs['caldb'] = caldb
        obs['irf']   = irf

        # Append observation
        obsdef.append(obs)

        # Update start time for next pointing
        tmin = set_tmin_for_next_pointing(tmin, duration)

    # Dump statistics
    print('Number of pointings: %d (%.2f s)' % (n_pnt,duration))

    # Return observation definition
    return obsdef


# ======================================= #
# Write observation definition dictionary #
# ======================================= #
def write_obsdef(filename, obsdef, idstart):
    """
    Write observation definition file

    Parameters
    ----------
    filename : str
        Observation definition file name
    obsdef : list of dict
        List of pointing definitions
    idstart : int
        First identifier of observation definition file
    """
    # Open file
    f = open(filename, 'w')

    # Write header
    f.write('id,ra,dec,tmin,duration,caldb,irf,emin,emax\n')

    # Initialise identifier
    obsid = idstart

    # Loop over pointings
    for obs in obsdef:

        # If we have lon,lat then convert into RA,Dec
        if 'lon' in obs and 'lat' in obs:
            lon = obs['lon']
            lat = obs['lat']
            pnt = gammalib.GSkyDir()
            pnt.lb_deg(lon,lat)
            ra  = pnt.ra_deg()
            dec = pnt.dec_deg()
        else:
            ra  = obs['ra']
            dec = obs['dec']

        # Set site dependent energy thresholds
        if 'South' in obs['irf']:
            emin =   0.030
            emax = 160.0
        else:
            emin =  0.030
            emax = 50.0

        # Write information
        f.write('%6.6d,%8.4f,%8.4f,%.4f,%.4f,%s,%s,%.3f,%.1f\n' %
                (obsid, ra, dec, obs['tmin'], obs['duration'], obs['caldb'], \
                 obs['irf'], emin, emax))

        # Increment identifier
        obsid += 1

    # Close file
    f.close()

    # Return
    return


# ============== #
# Make pointings #
# ============== #
def make_pointings():
    """
    Make pointings
    """
    # Initialise flags
    need_help = False

    # Test for command line arguments
    if (len(sys.argv) > 1):
        if sys.argv[1] == '-h':
            need_help = True
        else:
            obsname = sys.argv[1]
            if len(sys.argv) > 2:
                caldb = sys.argv[2]
            else:
                caldb = 'prod3b'
    else:
        need_help = True

    # Print help if needed and exit
    if need_help:
        print('Usage: make_pointing.py [OPTIONS] [CALDB]')
        print('     -h     Display this usage message')
        print('     gps    Galactic plane survey (2 row scheme)')
        print('     gps3   Galactic plane survey (3 row scheme)')
        print('     egal   Extragalactic survey')
        print('     gc     Galactic centre survey')
        print('     lmc    LMC survey')
        sys.exit()

    # Set LST simulation flag to true
    lst = True

    # Galactic plane survey
    if obsname == 'gps':
        obsdef = set_gps(caldb=caldb, lst=lst)
        write_obsdef('gps.dat', obsdef, 110000)
    elif obsname == 'gps_south':
        obsdef = set_gps(caldb=caldb, lst=lst, north=False)
        write_obsdef('gps.dat', obsdef, 120000)
    elif obsname == 'gps3':
        obsdef = set_gps(separation=1.5, caldb=caldb, lst=lst)
        write_obsdef('gps3.dat', obsdef, 130000)

    # Extragalactic survey
    elif obsname == 'egal':
        obsdef = set_egal(caldb=caldb, lst=lst)
        write_obsdef('egal.dat', obsdef, 210000)
    elif obsname == 'egal_south':
        obsdef = set_egal(caldb=caldb, lst=lst, north=False)
        write_obsdef('egal.dat', obsdef, 220000)

    # Galactic centre
    elif obsname == 'gc':
        obsdef = set_gc(caldb=caldb, lst=lst)
        write_obsdef('gc.dat', obsdef, 310000)

    # LMC
    elif obsname == 'lmc':
        obsdef = set_lmc(caldb=caldb, lst=lst)
        write_obsdef('lmc.dat', obsdef, 410000)

    # Invalid pattern
    else:
        print('Unknown option "' + obsname + '"')

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Make pointings
    make_pointings()



#! /usr/bin/env python
# ==========================================================================
# Puts subset of sources in diffuse model cube
#
# Copyright (C) 2017-2018 Josh Cardenzana
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools


# ================= #
# csmodelsois class #
# ================= #
class csmodelsois(ctools.cscript):
    """
    Puts subset of sources in diffuse model cube

    The csmodelsois class puts a subset of sources in a model definition
    XML file into a model cube and generates a new model definition XML file
    in which the subset of sources is replaced by a model of type
    DiffuseMapCube.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        
        Parameters
        ----------
        argv : list of parameters
        
        Raises
        ------
        TypeError
            An invalid number of command line arguments was provided.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)
        
        # Initialize parameters
        self._cubegen    = ctools.ctmapcube()
        self._models     = gammalib.GModels()
        self._cubemodels = gammalib.GModels()
        
        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Query input parameters
        self['inmodel'].filename()
        self['soilist'].string()

        # Read input parameters
        ptsrcsig = self['ptsrcsig'].real()

        # Get the cube centre coordinates
        self._cubegen['coordsys'].string(self['coordsys'].string())
        if self['coordsys'].string() == 'CEL':
            self._cubegen['xref'].real(self['ra'].real())
            self._cubegen['yref'].real(self['dec'].real())
        else:
            self._cubegen['xref'].real(self['glon'].real())
            self._cubegen['yref'].real(self['glat'].real())

        # Get spatial binning parameters
        self._cubegen['binsz'].real(self['binsz'].real())
        self._cubegen['nxpix'].integer(self['nxpix'].integer())
        self._cubegen['nypix'].integer(self['nypix'].integer())
        self._cubegen['proj'].string(self['proj'].string())

        # Get the energy binning parameters
        self._cubegen['ebinalg'].string(self['ebinalg'].string())
        if self['ebinalg'].string() == 'FILE':
            self._cubegen['ebinfile'].filename(self['ebinfile'].filename())
        else:
            self._cubegen['emin'].real(self['emin'].real())
            self._cubegen['emax'].real(self['emax'].real())
            self._cubegen['enumbins'].integer(self['enumbins'].integer())

        # Set point source significance
        self._cubegen['ptsrcsig'].real(ptsrcsig)

        # Read optionally output cube filenames
        if self._read_ahead():
            self._cubegen['outcube'].filename(self['outcube'].filename())
            self['outmodel'].query()

        # Write parameters into logger
        self._log_parameters(gammalib.TERSE)

        return

    def _gen_cubemodel(self):
        """
        Generates a binned model from the model file using ctmapcube
        """
        # Get a list of sources we DONT want to put into the cube
        sources = self['soilist'].string().split(',')

        # Load the model into a model container
        if (self._models.size() == 0):
            self._models = gammalib.GModels(self['inmodel'].filename())

        # Store a copy of the models
        self._cubemodels = self._models.copy()

        # Loop through models and pull out all models not in the list
        for model in self._models:

            # If model is not a sky model then continue
            if model.classname() != 'GModelSky':
                self._cubemodels.remove(model.name())

            # ... otherwise, if model is in list of sources the remove it
            # from the container
            elif model.name() in sources:
                self._cubemodels.remove(model.name())

        # Log the number of models to be put into the cube
        self._log_value(self['chatter'].integer(), 'Numner of cube models',
                        self._cubemodels.size())

        # Return
        return


    # Public methods
    def run(self):
        """
        Implements the actual bulk of the script's tasks
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Set the models to put into the cube
        self._gen_cubemodel()

        # Only run if there still models to be filled into the cube
        if not self._cubemodels.is_empty():

            # Log action
            self._log_string(gammalib.NORMAL, 'Generating model cube')

            # Fill the cube models into the ctmapcube object
            self._cubegen.models(self._cubemodels)

            # Run the map generating method
            self._cubegen.run()

        else:
            self._log_string(gammalib.NORMAL,
                             'List of models is empty, nothing will be done')

        # Return
        return

    def save(self):
        """
        Save the cube and optionally the updated XML model file
        """
        # Save the generated cube if the cube and filename are not empty
        if ((not self._cubegen.mapcube().cube().is_empty()) and
            self['outcube'].is_valid()):
            self._cubegen.mapcube().save(self['outcube'].filename(),
                                         self['clobber'].boolean())
        
        # If requested, save the updated list of models
        if self['outmodel'].is_valid():

            # Generate a list of models that will be output to file
            outmodels = gammalib.GModels(self._models)
            
            # Remove all models used in the generated cube
            for model in self._cubemodels:
                outmodels.remove(model.name())

            # Generate the actual cube model
            spat     = gammalib.GModelSpatialDiffuseCube(self['outcube'].filename())
            spec     = gammalib.GModelSpectralConst()
            newmodel = gammalib.GModelSky(spat, spec)

            # Give the model a default name
            newmodel.name(self.cubemodelname())

            # Now append the model to the list of output models
            outmodels.append(newmodel)

            # Save the model list
            outmodels.save(self['outmodel'].filename())

        # Return
        return

    def mapcube(self):
        """
        Return the mapcube generated by the underlying 'ctmapcube' object
        """
        # Return
        return self._cubegen.mapcube()

    def cubemodelname(self):
        """
        Return the name to be given to the cube model in the output XML file
        """
        # Return
        return 'FixedSourcesCube'

    def models(self, models):
        """
        Set the models

        Parameters
        -------
        models : `~gammalib.GModels`
            Model container
        """
        # Set models
        self._models = models

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':
    
    # Create instance of application
    app = csmodelsois(sys.argv)
    
    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctskymap tool.
#
# Copyright (C) 2014-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ============================ #
# Test class for ctskymap tool #
# ============================ #
class Test(test):
    """
    Test class for ctskymap tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._skymap_ring      = self._datadir + '/skymap_ring.fits'
        self._exclusion_region = self._datadir + '/exclusion.reg'
        self._exclusion_map    = self._datadir + '/exclusion.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctskymap')

        # Append tests
        self.append(self._test_cmd, 'Test ctskymap on command line')
        self.append(self._test_python, 'Test ctskymap from Python')
        self.append(self._test_python_irf,
                    'Test ctskymap from Python for IRF background')
        self.append(self._test_python_ring,
                    'Test ctskymap from Python for RING background')
        self.append(self._test_python_exclusion_map,
                    'Test ctskymap exclusion map methods from Python')

        # Return
        return

    # Test ctskymap on command line
    def _test_cmd(self):
        """
        Test ctskymap on the command line
        """
        # Set tool name
        ctskymap = self._tool('ctskymap')

        # Setup ctskymap command
        cmd = ctskymap+' inobs="'+self._events+'"'+\
                       ' outmap="ctskymap_cmd1.fits"'+ \
                       ' emin=0.1 emax=100.0 nxpix=20 nypix=20'+ \
                       ' binsz=0.2 coordsys="CEL" proj="CAR"'+ \
                       ' xref=83.63 yref=22.01 bkgsubtract="NONE"'+ \
                       ' logfile="ctskymap_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctskymap_cmd1.fits')

        # Setup ctskymap command
        cmd = ctskymap+' inobs="event_file_that_does_not_exist.fits"'+\
                       ' outmap="ctskymap_cmd2.fits"'+ \
                       ' emin=0.1 emax=100.0 nxpix=20 nypix=20'+ \
                       ' binsz=0.2 coordsys="CEL" proj="CAR"'+ \
                       ' xref=83.63 yref=22.01 bkgsubtract="NONE"'+ \
                       ' logfile="ctskymap_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctskymap --help
        self._check_help(ctskymap)

        # Return
        return

    # Test ctskymap from Python
    def _test_python(self):
        """
        Test ctskymap from Python
        """
        # Allocate ctskymap
        skymap = ctools.ctskymap()

        # Check that empty ctskymap tool holds a map that has no pixels
        self._check_map(skymap.skymap(), nx=0, ny=0)

        # Check that saving does not nothing
        skymap['outmap']  = 'ctskymap_py0.fits'
        skymap['logfile'] = 'ctskymap_py0.log'
        skymap.logFileOpen()
        skymap.save()
        self.test_assert(not os.path.isfile('ctskymap_py0.fits'),
             'Check that no sky map has been created')

        # Check that clearing does not lead to an exception or segfault
        skymap.clear()

        # Now set ctskymap parameters
        skymap['inobs']       = self._events
        skymap['emin']        = 0.1
        skymap['emax']        = 100
        skymap['nxpix']       = 20
        skymap['nypix']       = 20
        skymap['binsz']       = 0.2
        skymap['coordsys']    = 'CEL'
        skymap['proj']        = 'CAR'
        skymap['xref']        = 83.63
        skymap['yref']        = 22.01
        skymap['bkgsubtract'] = 'NONE'
        skymap['outmap']      = 'ctskymap_py1.fits'
        skymap['logfile']     = 'ctskymap_py1.log'
        skymap['chatter']     = 2

        # Run ctskymap tool
        skymap.logFileOpen()   # Make sure we get a log file
        skymap.run()
        skymap.save()

        # Check result file
        self._check_result_file('ctskymap_py1.fits')

        # Copy ctskymap tool
        cpy_skymap = skymap.copy()

        # Check sky map of ctskymap copy
        self._check_map(cpy_skymap.skymap())

        # Execute copy of ctskymap tool again, now with a higher chatter
        # level than before
        cpy_skymap['usepnt']   = True
        cpy_skymap['emin']     = 0.2
        cpy_skymap['emax']     = 150.0
        cpy_skymap['coordsys'] = 'GAL'
        cpy_skymap['proj']     = 'CAR'
        cpy_skymap['outmap']   = 'ctskymap_py2.fits'
        cpy_skymap['logfile']  = 'ctskymap_py2.log'
        cpy_skymap['publish']  = True
        cpy_skymap['chatter']  = 3
        cpy_skymap.logFileOpen()  # Needed to get a new log file
        cpy_skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_py2.fits')

        # Now clear copy of ctskymap tool
        cpy_skymap.clear()

        # Check that cleared ctskymap tool holds a map that has no pixels
        self._check_map(cpy_skymap.skymap(), nx=0, ny=0)

        # Get mixed observation container
        obs = self._obs_mixed()

        # Allocate ctskymap tool from observation container
        skymap = ctools.ctskymap(obs)
        skymap['emin']        = 0.1
        skymap['emax']        = 100
        skymap['nxpix']       = 20
        skymap['nypix']       = 20
        skymap['binsz']       = 0.2
        skymap['coordsys']    = 'GAL'
        skymap['proj']        = 'CAR'
        skymap['xref']        = 184.5575
        skymap['yref']        =  -5.7844
        skymap['bkgsubtract'] = 'NONE'
        skymap['outmap']      = 'ctskymap_py3.fits'
        skymap['logfile']     = 'ctskymap_py3.log'
        skymap['chatter']     = 4

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_py3.fits')

        # Publish with name
        skymap.publish('My sky map')

        # Return
        return

    # Test ctskymap from Python for IRF background method
    def _test_python_irf(self):
        """
        Test ctskymap from Python for IRF background method
        """
        # Test IRF background subtraction
        skymap = ctools.ctskymap()
        skymap['inobs']       = self._events
        skymap['emin']        = 0.1
        skymap['emax']        = 100
        skymap['nxpix']       = 20
        skymap['nypix']       = 20
        skymap['binsz']       = 0.2
        skymap['coordsys']    = 'GAL'
        skymap['proj']        = 'CAR'
        skymap['xref']        = 184.5575
        skymap['yref']        =  -5.7844
        skymap['bkgsubtract'] = 'IRF'
        skymap['caldb']       = self._caldb
        skymap['irf']         = self._irf
        skymap['outmap']      = 'ctskymap_irf_py1.fits'
        skymap['logfile']     = 'ctskymap_irf_py1.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_irf_py1.fits')
        self._check_result_file('ctskymap_irf_py1.fits[BACKGROUND]')
        self._check_result_file('ctskymap_irf_py1.fits[SIGNIFICANCE]')

        # Return
        return

    # Test ctskymap from Python for RING background method
    def _test_python_ring(self):
        """
        Test ctskymap from Python for RING background method
        """
        # Test basic RING background subtraction
        skymap = ctools.ctskymap()
        skymap['inobs']       = self._events
        skymap['emin']        = 0.1
        skymap['emax']        = 100
        skymap['nxpix']       = 10
        skymap['nypix']       = 10
        skymap['binsz']       = 0.2
        skymap['coordsys']    = 'CEL'
        skymap['proj']        = 'CAR'
        skymap['xref']        = 83.63
        skymap['yref']        = 22.01
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 0
        skymap['inexclusion'] = 'NONE'
        skymap['caldb']       = self._caldb
        skymap['irf']         = self._irf
        skymap['outmap']      = 'ctskymap_ring_py1.fits'
        skymap['logfile']     = 'ctskymap_ring_py1.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py1.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py1.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py1.fits[SIGNIFICANCE]', 10, 10)

        # Check generated fits object
        self._check_map(gammalib.GSkyMap(skymap.fits().image('SKYMAP')), nx=10, ny=10)
        self._check_map(gammalib.GSkyMap(skymap.fits().image('BACKGROUND')), nx=10, ny=10)
        self._check_map(gammalib.GSkyMap(skymap.fits().image('SIGNIFICANCE')), nx=10, ny=10)

        # Test iterative RING background subtraction
        skymap = ctools.ctskymap()
        skymap['inobs']       = self._events
        skymap['emin']        = 0.1
        skymap['emax']        = 100
        skymap['nxpix']       = 10
        skymap['nypix']       = 10
        skymap['binsz']       = 0.2
        skymap['coordsys']    = 'CEL'
        skymap['proj']        = 'CAR'
        skymap['xref']        = 83.63
        skymap['yref']        = 22.01
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 2
        skymap['threshold']   = 5.0
        skymap['inexclusion'] = 'NONE'
        skymap['caldb']       = self._caldb
        skymap['irf']         = self._irf
        skymap['outmap']      = 'ctskymap_ring_py2.fits'
        skymap['logfile']     = 'ctskymap_ring_py2.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py2.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py2.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py2.fits[SIGNIFICANCE]', 10, 10)

        # Test iterative RING background subtraction, starting from a sky map
        skymap = ctools.ctskymap()
        skymap['inmap']       = self._skymap_ring
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 2
        skymap['threshold']   = 5.0
        skymap['inexclusion'] = 'NONE'
        skymap['outmap']      = 'ctskymap_ring_py3.fits'
        skymap['logfile']     = 'ctskymap_ring_py3.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py3.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py3.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py3.fits[SIGNIFICANCE]', 10, 10)

        # Test iterative RING background subtraction with an exclusion region
        skymap = ctools.ctskymap()
        skymap['inmap']       = self._skymap_ring
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 2
        skymap['threshold']   = 5.0
        skymap['inexclusion'] = self._exclusion_region
        skymap['outmap']      = 'ctskymap_ring_py4.fits'
        skymap['logfile']     = 'ctskymap_ring_py4.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py4.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py4.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py4.fits[SIGNIFICANCE]', 10, 10)

        # Test iterative RING background subtraction with an exclusion map
        skymap = ctools.ctskymap()
        skymap['inmap']       = self._skymap_ring
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 2
        skymap['threshold']   = 5.0
        skymap['inexclusion'] = self._exclusion_map
        skymap['outmap']      = 'ctskymap_ring_py5.fits'
        skymap['logfile']     = 'ctskymap_ring_py5.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py5.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py5.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py5.fits[SIGNIFICANCE]', 10, 10)

        # Test RING background subtraction with direct significance computation
        skymap = ctools.ctskymap()
        skymap['inmap']       = self._skymap_ring
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 0
        skymap['threshold']   = 5.0
        skymap['inexclusion'] = 'NONE'
        skymap['usefft']      = False  # Direct computation, no FFT
        skymap['outmap']      = 'ctskymap_ring_py6.fits'
        skymap['logfile']     = 'ctskymap_ring_py6.log'
        skymap['chatter']     = 2

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py6.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py6.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py6.fits[SIGNIFICANCE]', 10, 10)

        # Return
        return

    # Test exclusion_map methods
    def _test_python_exclusion_map(self):
        """
        Test exclusion_map methods
        """
        # Setup ctskymap
        skymap = ctools.ctskymap()

        # Set exclusion map
        map = gammalib.GSkyRegionMap(self._exclusion_map)

        # Test iterative RING background subtraction with an exclusion map
        # set via the ctskymap::exclusion_map() method
        skymap = ctools.ctskymap()
        skymap['inmap']       = self._skymap_ring
        skymap['bkgsubtract'] = 'RING'
        skymap['roiradius']   = 0.1
        skymap['inradius']    = 0.6
        skymap['outradius']   = 0.8
        skymap['iterations']  = 2
        skymap['threshold']   = 5.0
        skymap['outmap']      = 'ctskymap_ring_py7.fits'
        skymap['logfile']     = 'ctskymap_ring_py7.log'
        skymap['chatter']     = 2
        skymap.exclusion_map(map)

        # Execute tool
        skymap.logFileOpen()  # Needed to get a new log file
        skymap.execute()

        # Check result file
        self._check_result_file('ctskymap_ring_py7.fits', 10, 10)
        self._check_result_file('ctskymap_ring_py7.fits[BACKGROUND]', 10, 10)
        self._check_result_file('ctskymap_ring_py7.fits[SIGNIFICANCE]', 10, 10)

        # Recover exclusion map
        map2 = skymap.exclusion_map()

        # Check exclusion map
        self._check_map(map2.map(), nx=10, ny=10)

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nx=20, ny=20):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Sky map file name
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Open result file
        skymap = gammalib.GSkyMap(filename)

        # Check sky map
        self._check_map(skymap, nx=nx, ny=ny)

        # Return
        return

    # Check sky map
    def _check_map(self, skymap, nx=20, ny=20):
        """
        Check sky map

        Parameters
        ----------
        skymap : `~gammalib.GSkyMap`
            Sky map
        nx : int, optional
            Number of X pixels
        ny : int, optional
            Number of Y pixels
        """
        # Determine number of maps
        if nx > 0 and ny > 0:
            nmaps = 1
        else:
            nmaps = 0

        # Check dimensions
        self.test_value(skymap.nmaps(), nmaps, 'Check number of maps')
        self.test_value(skymap.nx(), nx, 'Check for number of X pixels')
        self.test_value(skymap.ny(), ny, 'Check for number of Y pixels')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Display spectrum generated by csspec
#
# Copyright (C) 2015-2020 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()
try:
    import numpy as np
except ImportError:
    print('This script needs the "numpy" module')
    sys.exit()


# ===================================== #
# Extract the spectrum info from a file #
# ===================================== #
def get_spectrum_file(filename):
    """
    Extract the spectrum info from a file for plotting

    Parameters
    ----------
    filename : str
        Name of spectrum FITS file

    Returns
    -------
    spec : dict
        Python dictionary defining spectral plot parameters
    """
    # Read spectrum file    
    fits = gammalib.GFits(filename)

    # Return dictionary
    return get_spectrum_fits(fits)


# ============================================= #
# Extract the spectrum info from a GFits object #
# ============================================= #
def get_spectrum_fits(fits):
    """
    Extract the spectrum info from a GFits object

    Parameters
    ----------
    fits : `~gammalib.GFits`
        Spectral GFits object
    
    Returns
    -------
    spec : dict
        Python dictionary defining spectral plot parameters
    """
    # Read spectrum objects
    table    = fits.table(1)
    c_energy = table['e_ref']
    c_ed     = table['e_min']
    c_eu     = table['e_max']
    c_flux   = table['ref_e2dnde']
    c_norm   = table['norm']
    c_eflux  = table['norm_err']
    c_upper  = table['norm_ul']
    c_ts     = table['ts']

    # Initialise arrays to be filled
    spec = {
        'energies'    : [],
        'flux'        : [],
        'ed_engs'     : [],
        'eu_engs'     : [],
        'e_flux'      : [],
        'ul_energies' : [],
        'ul_ed_engs'  : [],
        'ul_eu_engs'  : [],
        'ul_flux'     : [],
        'yerr'        : [],
        'loglike'     : [],
        'engs_scan'   : [],
        'e2dnde_scan' : [],
        'dll_scan'    : []
    }

    # Determine if we can load the delta-log-likelihood profiles
    has_sedtype = table.has_card('SED_TYPE')
    load_dll    = False
    if has_sedtype:
        seds = table.card('SED_TYPE').string().split(',')

        # Load likelhood columns if present
        if seds[0] == 'likelihood':
            load_dll    = True
            c_loglike   = table['loglike']
            c_norm_scan = table['norm_scan']
            c_dll_scan  = table['dloglike_scan']

    # Loop over rows of the file
    nrows = table.nrows()
    for row in range(nrows):

        # Get Test Statistic, flux and flux error
        ts    = c_ts.real(row)
        norm  = c_norm.real(row)
        flx   = norm * c_flux.real(row)
        e_flx = flx * c_eflux.real(row)

        # If Test Statistic is larger than 9 and flux error is smaller than
        # flux then append flux plots ...
        if ts > 9.0 and e_flx < flx:
            spec['energies'].append(c_energy.real(row))
            spec['flux'].append(flx)
            spec['ed_engs'].append(c_ed.real(row))
            spec['eu_engs'].append(c_eu.real(row))
            spec['e_flux'].append(e_flx)

        # ... otherwise append upper limit
        else:
            spec['ul_energies'].append(c_energy.real(row))
            spec['ul_flux'].append(flx*c_upper.real(row))
            spec['ul_ed_engs'].append(c_ed.real(row))
            spec['ul_eu_engs'].append(c_eu.real(row))

    # Load the delta log-likelihood values
    if load_dll:
        spec['e2dnde_scan'] = np.zeros((nrows, c_norm_scan.elements(0)))
        spec['dll_scan']    = np.zeros((nrows, c_norm_scan.elements(0)))

        for row in range(nrows):
            loglike = c_loglike.real(row)
            spec['loglike'].append(loglike)
            spec['engs_scan'].append(c_energy.real(row) - c_ed.real(row))
            for col in range(c_norm_scan.elements(row)):
                spec['e2dnde_scan'][row,col] = c_norm_scan.real(row,col) * c_flux.real(row)
                spec['dll_scan'][row,col] = c_dll_scan.real(row,col)
        spec['engs_scan'].append(c_energy.real(nrows-1)+c_eu.real(nrows-1))

    # Set upper limit errors
    spec['yerr'] = [0.6 * x for x in spec['ul_flux']]

    # Return dictionary
    return spec


# ========================= #
# Plot delta log-likelihood #
# ========================= #
def plot_dloglike(spec):
    """
    Plot delta log-likelihood

    Parameters
    ----------
    spec : dict
        Python dictionary defining spectral plot parameters
    """
    # Establish the bounds on the x,y plot
    ymin,ymax = plt.gca().get_ylim()
    ymin      = np.log10(ymin)
    ymax      = np.log10(ymax)
    steps     = 1000
    ebins     = len(spec['e2dnde_scan'])
    y_bounds  = np.linspace(ymin, ymax, steps+1)

    # Compute the logarithmic center of each bin
    fluxpnts = 10 ** ((y_bounds[1:]+y_bounds[:-1]) / 2.0)
    
    # Scale the boundaries on the grid
    y_bounds = 10 ** y_bounds

    # Interpolate the dlogLike values
    dll_hist = [[]]*ebins
    for ebin in range(ebins):
        dll_hist[ebin] = np.interp(fluxpnts, spec['e2dnde_scan'][ebin], 
                                             spec['dll_scan'][ebin])

    # Transpose the dll_hist array so x=energy, y=dnde
    dll_hist = [[row[i] for row in dll_hist] for i in range(steps)]

    # Plot the likelihood profiles
    plt.pcolormesh(np.array(spec['engs_scan']), y_bounds,
                   np.array(dll_hist), vmin=-10.0, vmax=0.0,
                   cmap='Reds', alpha=0.9)
    cbar = plt.colorbar()
    cbar.set_label(r'$\Delta$ log-likelihood')
    plt.grid()

    # Return
    return


# ============= #
# Plot spectrum #
# ============= #
def plot_spectrum(filename, plotfile):
    """
    Plot spectrum

    Parameters
    ----------
    filename : str
        Name of spectrum FITS file
    plotfile : str
        Plot file name
    """
    # Get spectrum parameters
    spec = get_spectrum_file(filename)

    # Create the plot
    plt.figure()
    plt.loglog()
    plt.grid()
    plt.xlabel('Energy (TeV)')
    plt.ylabel(r'E$^2$ $\times$ dN/dE (erg cm$^{-2}$ s$^{-1}$)')

    # Plot the spectrum
    plt.errorbar(spec['energies'], spec['flux'], 
                 yerr=spec['e_flux'], xerr=[spec['ed_engs'], spec['eu_engs']],
                 fmt='ro')

    # Plot upper limits
    if len(spec['ul_energies'])	> 0:
        plt.errorbar(spec['ul_energies'], spec['ul_flux'], yerr=spec['yerr'],
                     xerr=[spec['ul_ed_engs'], spec['ul_eu_engs']],
                     uplims=True, fmt='ro')

    # Plot log-likelihood profiles
    if len(spec['dll_scan']) > 0:
        plot_dloglike(spec)

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ============= #
# Show spectrum #
# ============= #
def show_spectrum():
    """
    Show spectrum
    """
    # Set usage string
    usage = 'show_spectrum.py [-p plotfile] [file]'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Show spectrum
    plot_spectrum(args[0], plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show spectrum
    show_spectrum()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csobsselect script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================= #
# Test class for csobsselect script #
# ================================= #
class Test(test):
    """
    Test class for csobsselect script

    This test class makes unit tests for the csobsselect script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor.
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csobsselect')

        # Append tests
        self.append(self._test_cmd, 'Test csobsselect on command line')
        self.append(self._test_python, 'Test csobsselect from Python')

        # Return
        return

    # Test csobsselect on command line
    def _test_cmd(self):
        """
        Test csobsselect on the command line
        """
        # Set script name
        csobsselect = self._script('csobsselect')

        # Setup csobsselect command
        cmd = csobsselect+' inobs="'+self._events+'"'+ \
                          ' outobs="csobsselect_cmd1.xml"'+ \
                          ' pntselect="CIRCLE" coordsys="CEL" ra=83.63'+ \
                          ' dec=22.01 rad=5.0 tmin=NONE tmax=NONE'+ \
                          ' logfile="csobsselect_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check observation definition file
        self._check_observation_file('csobsselect_cmd1.xml', 1)

        # Setup csobsselect command
        cmd = csobsselect+' inobs="observation_that_does_not_exist.xml"'+ \
                          ' outobs="csobsselect_cmd2.xml"'+ \
                          ' pntselect="CIRCLE" coordsys="CEL" ra=83.63'+ \
                          ' dec=22.01 rad=5.0 tmin=NONE tmax=NONE'+ \
                          ' logfile="csobsselect_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csobsselect --help
        self._check_help(csobsselect)

        # Return
        return

    # Test csobsselect from Python
    def _test_python(self):
        """
        Test csobsselect from Python
        """
        # Allocate empty csobsselect script
        obsselect = cscripts.csobsselect()

        # Check that saving does not nothing
        obsselect['outobs']  = 'csobsselect_py0.xml'
        obsselect['logfile'] = 'csobsselect_py0.log'
        obsselect.logFileOpen()
        obsselect.save()

        # Check than an empty observation definition file has been created
        self._check_observation_file('csobsselect_py0.xml', 0)

        # Check that clearing does not lead to an exception or segfault
        obsselect.clear()

        # Now set csobsselect parameters
        obsselect['inobs']     = self._events
        obsselect['pntselect'] = 'CIRCLE'
        obsselect['coordsys']  = 'CEL'
        obsselect['ra']        = 83.63
        obsselect['dec']       = 22.01
        obsselect['rad']       = 5.0
        obsselect['tmin']      = 'NONE'
        obsselect['tmax']      = 'NONE'
        obsselect['outobs']    = 'csobsselect_py1.xml'
        obsselect['logfile']   = 'csobsselect_py1.log'
        obsselect['chatter']   = 2

        # Run csobsselect script and save observations
        obsselect.logFileOpen()   # Make sure we get a log file
        obsselect.run()
        obsselect.save()

        # Check model file
        self._check_observation_file('csobsselect_py1.xml', 1)

        # Execute csobsselect script again, now in Galactic coordinates
        obsselect['pntselect'] = 'CIRCLE'
        obsselect['coordsys']  = 'GAL'
        obsselect['glon']      = 184.56
        obsselect['glat']      = -5.79
        obsselect['rad']       = 5.0
        obsselect['tmin']      = 'NONE'
        obsselect['tmax']      = 'NONE'
        obsselect['outobs']    = 'csobsselect_py2.xml'
        obsselect['logfile']   = 'csobsselect_py2.log'
        obsselect['chatter']   = 3
        obsselect.logFileOpen()  # Needed to get a new log file
        obsselect.execute()

        # Check model file
        self._check_observation_file('csobsselect_py2.xml', 1)

        # Execute csobsselect script again, now a box in Galactic coordinates
        obsselect['pntselect'] = 'BOX'
        obsselect['coordsys']  = 'GAL'
        obsselect['glon']      = 184.56
        obsselect['glat']      = -5.79
        obsselect['width']     = 5.0
        obsselect['height']    = 5.0
        obsselect['tmin']      = 'NONE'
        obsselect['tmax']      = 'NONE'
        obsselect['outobs']    = 'csobsselect_py3.xml'
        obsselect['logfile']   = 'csobsselect_py3.log'
        obsselect['chatter']   = 4
        obsselect.logFileOpen()  # Needed to get a new log file
        obsselect.execute()

        # Check model file
        self._check_observation_file('csobsselect_py3.xml', 1)

        # Execute csobsselect script again, now a box in celestial coordinates
        obsselect['pntselect'] = 'BOX'
        obsselect['coordsys']  = 'CEL'
        obsselect['ra']        = 83.63
        obsselect['dec']       = 22.01
        obsselect['width']     = 5.0
        obsselect['height']    = 5.0
        obsselect['tmin']      = 'NONE'
        obsselect['tmax']      = 'NONE'
        obsselect['outobs']    = 'csobsselect_py4.xml'
        obsselect['logfile']   = 'csobsselect_py4.log'
        obsselect['chatter']   = 4
        obsselect.logFileOpen()  # Needed to get a new log file
        obsselect.execute()

        # Check model file
        self._check_observation_file('csobsselect_py4.xml', 1)

        # Execute csobsselect script again, now with time selection
        obsselect['pntselect'] = 'BOX'
        obsselect['coordsys']  = 'CEL'
        obsselect['ra']        = 83.63
        obsselect['dec']       = 22.01
        obsselect['width']     = 5.0
        obsselect['height']    = 5.0
        obsselect['tmin']      = 0.0
        obsselect['tmax']      = 100.0
        obsselect['outobs']    = 'csobsselect_py5.xml'
        obsselect['logfile']   = 'csobsselect_py5.log'
        obsselect['chatter']   = 4
        obsselect.logFileOpen()  # Needed to get a new log file
        obsselect.execute()

        # Check model file
        self._check_observation_file('csobsselect_py5.xml', 0)

        # Return
        return

    # Check observation definition file
    def _check_observation_file(self, filename, number):
        """
        Check observation definition file
        """
        # Open observations
        obs = gammalib.GObservations(filename)

        # Check number of models
        self.test_value(obs.size(), number, 'Check for '+str(number)+
                        ' observations in XML file')
        
        # Return
        return
